from collections import OrderedDict

import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.nn import functional as F

from drvi.nn_modules.embedding import FeatureEmbedding


class MultiOneHotEncoding(nn.Module):
    def __init__(self, n_embedding_list: list[int], **kwargs):
        super().__init__()
        self.n_embedding_list = n_embedding_list
        self.device_container = nn.Parameter(torch.tensor([]))

    def forward(self, index_list: torch.Tensor) -> torch.Tensor:
        assert index_list.shape[-1] == len(self.n_embedding_list)
        result = torch.concat(
            [
                F.one_hot(index_list[..., i], num_classes=n_embedding)
                for i, n_embedding in enumerate(self.n_embedding_list)
            ],
            dim=-1,
        )
        return result.to(self.device_container.device)

    def get_extra_state(self):
        return {"n_embedding_list": self.n_embedding_list}

    def set_extra_state(self, state):
        self.n_embedding_list = state["n_embedding_list"]

    @property
    def embedding_dim(self):
        return sum(self.n_embedding_list)


class FeatureOneHotEncoding(FeatureEmbedding):
    def __init__(self, vocab_list: list[list[str]], **kwargs):
        n_vocab_list = [len(vocab) for vocab in vocab_list]
        super().__init__(vocab_list, n_vocab_list, **kwargs)

    @staticmethod
    def define_embeddings(n_embedding_list, embedding_dims, **kwargs):
        for n_key, dim in zip(n_embedding_list, embedding_dims, strict=False):
            assert n_key == dim
        return MultiOneHotEncoding(n_embedding_list, **kwargs)

    @classmethod
    def from_numpy_array(cls, sentences_array: np.ndarray, **kwargs):
        word_list = sentences_array.transpose().tolist()
        vocab_list = [list(OrderedDict.fromkeys(words)) for words in word_list]
        return cls(vocab_list, **kwargs)

    @classmethod
    def from_pandas_dataframe(cls, sentences_df: pd.DataFrame, **kwargs):
        return cls.from_numpy_array(sentences_df.values)



from collections import defaultdict
from uuid import uuid4

import scvi
from scvi._types import AnnOrMuData
from scvi.data import AnnDataManager, AnnDataManagerValidationCheck, _constants

from drvi.scvi_tools_based.merlin_data._data import MerlinData
from drvi.scvi_tools_based.merlin_data.fields import (
    MerlinCategoricalJointObsField,
    MerlinCategoricalObsField,
    MerlinDataField,
    MerlinLayerField,
    MerlinNumericalJointObsField,
)


class MerlinDataManager(AnnDataManager):
    """
    Provides an interface to validate and process a MerlinData object for use in scvi-tools.

    Parameters
    ----------
    fields : list[MerlinDataField] | None, optional
        List of MerlinDataField objects to register.
    setup_method_args : dict | None, optional
        Dictionary describing the model and arguments passed in by the user
        to setup this AnnDataManager.
    validation_checks : AnnDataManagerValidationCheck | None, optional
        DataClass specifying which global validation checks to run on the data object.
    """

    def __init__(
        self,
        fields: list[MerlinDataField] | None = None,
        setup_method_args: dict | None = None,
        validation_checks: AnnDataManagerValidationCheck | None = None,
    ) -> None:
        self.id = str(uuid4())
        self.adata = None
        self.fields = fields or []
        self.validation_checks = validation_checks or AnnDataManagerValidationCheck()
        self._registry = {
            _constants._SCVI_VERSION_KEY: scvi.__version__,
            _constants._MODEL_NAME_KEY: None,
            _constants._SETUP_ARGS_KEY: None,
            _constants._FIELD_REGISTRIES_KEY: defaultdict(dict),
        }
        if setup_method_args is not None:
            self._registry.update(setup_method_args)

    def _validate_anndata_object(self, adata: AnnOrMuData | MerlinData):
        """For a given AnnData object, runs general scvi-tools compatibility checks."""
        if isinstance(adata, MerlinData):
            return True
        return super()._validate_anndata_object(adata)

    def get_fields_schema_mapping(self):
        mapping = []
        for field in self.fields:
            if isinstance(field, MerlinLayerField):
                assert field.attr_name == _constants._ADATA_ATTRS.LAYERS
                mapping.append((field.registry_key, field.attr_key))
            elif isinstance(field, MerlinCategoricalObsField):
                if field.is_default:
                    continue
                assert field.attr_name == _constants._ADATA_ATTRS.OBS
                mapping.append((field.registry_key, field._original_attr_key))
            elif isinstance(field, MerlinCategoricalJointObsField):
                assert field.attr_name == _constants._ADATA_ATTRS.OBSM
                if len(field.attr_keys) > 0:
                    mapping.append((field.registry_key, field.attr_keys))
            elif isinstance(field, MerlinNumericalJointObsField):
                assert field.attr_name == _constants._ADATA_ATTRS.OBSM
                if len(field.attr_keys) > 0:
                    mapping.append((field.registry_key, field.attr_keys))
            else:
                raise NotImplementedError()
        return mapping

    def get_dataset(self, split, **dataset_kwargs):
        """
        Get the dataset for a given split.

        Parameters
        ----------
            split (str): The split to retrieve the dataset for.
            **dataset_kwargs: Additional keyword arguments to pass to the dataset.

        Returns
        -------
            The dataset for the given split.
        """
        merlin_data = self.adata
        columns = []
        for _, col in self.get_fields_schema_mapping():
            if isinstance(col, list):
                columns.extend(col)
            else:
                columns.append(col)
        return merlin_data.get_dataset(split, columns, **dataset_kwargs)



from torch import nn


class SimpleResidual(nn.Module):
    def __init__(self, layer):
        super().__init__()
        self.layer = layer

    def forward(self, x):
        return x + self.layer(x)



import gc

import torch
from merlin.dataloader.torch import Loader


class MerlinTransformedDataLoader(Loader):
    """
    Data loader for MerlinData

    Parameters
    ----------
    mapping : list of tuple specifying (target_col, source_col)
    """

    def __init__(self, *args, mapping=None, gc_every_n_iter=100, **kwargs):
        super().__init__(*args, **kwargs)
        self.mapping = mapping
        self.gc_every_n_iter = gc_every_n_iter
        self.iters_to_gc = gc_every_n_iter

    def convert_batch(self, batch):
        batch = super().convert_batch(batch)
        batch, _ = batch
        self.iters_to_gc -= 1
        if self.iters_to_gc < 1:
            self.iters_to_gc = self.gc_every_n_iter
            gc.collect()
        if self.mapping is None:
            return batch
        result = {}
        for new_col, col in self.mapping:
            if isinstance(col, list):
                result[new_col] = torch.stack([batch[c] for c in col], dim=1)
                for c in col:
                    del batch[c]
            else:
                result[new_col] = batch[col]
        return result



import logging
from collections.abc import Callable, Sequence

import numpy as np
import scvi
import torch
from anndata import AnnData
from scvi import REGISTRY_KEYS
from torch.nn import functional as F

logger = logging.getLogger(__name__)


class GenerativeMixin:
    """Mixins to interpret generative part of the method."""

    @torch.inference_mode()
    def iterate_on_decoded_latent_samples(
        self,
        z: np.ndarray,
        step_func: Callable,
        aggregation_func: Callable,
        lib: np.ndarray | None = None,
        cat_values: np.ndarray | None = None,
        cont_values: np.ndarray | None = None,
        batch_size=scvi.settings.batch_size,
        map_cat_values: bool = False,
    ) -> np.ndarray:
        r"""Iterate over decoder outputs and aggregate the results.

        Parameters
        ----------
        z
            Latent samples.
        step_func
            Function to apply to the decoder output at each step.
            generative_outputs and a store variable are given to the function
        aggregation_func
            Function to aggregate the step results.
        lib
            Library size array.
        cat_values
            Categorical covariates (required if model has categorical key).
        cont_values
            Continuous covariates.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        map_cat_values
            Whether to map categorical covariates in cat_values to integers based on anndata manager pipeline
        """
        store = []
        self.module.eval()

        if cat_values is not None and map_cat_values:
            if cat_values.ndim == 1:  # For a user not noticing cat_values should be 2d!
                cat_values = cat_values.reshape(-1, 1)
            mapped_values = np.zeros_like(cat_values)
            for i, (_label, map_keys) in enumerate(
                self.adata_manager.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY)["mappings"].items()
            ):
                cat_mapping = dict(zip(map_keys, range(len(map_keys)), strict=False))
                mapped_values[:, i] = np.vectorize(cat_mapping.get)(cat_values[:, i])
            cat_values = mapped_values.astype(np.int32)

        with torch.no_grad():
            for i in np.arange(0, z.shape[0], batch_size):
                slice = np.arange(i, min(i + batch_size, z.shape[0]))
                z_tensor = torch.tensor(z[slice])
                if lib is None:
                    lib_tensor = torch.tensor([1e4] * slice.shape[0])
                else:
                    lib_tensor = torch.tensor(lib[slice])
                cat_tensor = torch.tensor(cat_values[slice]) if cat_values is not None else None
                cont_tensor = torch.tensor(cont_values[slice]) if cont_values is not None else None
                batch_tensor = None

                gen_input = self.module._get_generative_input(
                    tensors={
                        REGISTRY_KEYS.BATCH_KEY: batch_tensor,
                        REGISTRY_KEYS.LABELS_KEY: None,
                        REGISTRY_KEYS.CONT_COVS_KEY: cont_tensor,
                        REGISTRY_KEYS.CAT_COVS_KEY: cat_tensor,
                    },
                    inference_outputs={
                        "z": z_tensor,
                        "library": lib_tensor,
                        "gene_likelihood_additional_info": {},
                    },
                )
                gen_output = self.module.generative(**gen_input)
                step_func(gen_output, store)
        result = aggregation_func(store)
        return result

    @torch.inference_mode()
    def decode_latent_samples(
        self,
        z: np.ndarray,
        lib: np.ndarray | None = None,
        cat_values: np.ndarray | None = None,
        cont_values: np.ndarray | None = None,
        batch_size=scvi.settings.batch_size,
        map_cat_values: bool = False,
    ) -> np.ndarray:
        r"""Return the distribution produces by the decoder for the given latent samples.

        This is typically considered as :math:`p(x \mid z)`.

        Parameters
        ----------
        z
            Latent samples.
        lib
            Library size array.
        cat_values
            Categorical covariates (required if model has categorical key).
        cont_values
            Continuous covariates.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        map_cat_valuse
            Whether to map categorical covariates in cat_values to integers based on anndata manager pipeline
        """
        step_func = lambda gen_output, store: store.append(gen_output["params"]["mean"].detach().cpu())
        aggregation_func = lambda store: torch.cat(store, dim=0).numpy(force=True)

        return self.iterate_on_decoded_latent_samples(
            z=z,
            step_func=step_func,
            aggregation_func=aggregation_func,
            lib=lib,
            cat_values=cat_values,
            cont_values=cont_values,
            batch_size=batch_size,
            map_cat_values=map_cat_values,
        )

    @torch.inference_mode()
    def iterate_on_ae_output(
        self,
        adata: AnnData,
        step_func: Callable,
        aggregation_func: Callable,
        indices: Sequence[int] | None = None,
        batch_size: int | None = None,
        deterministic: bool = False,
    ) -> np.ndarray:
        r"""Iterate over autoencoder outputs and aggregate the results.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData.
            If `None`, defaults to the AnnData object used to initialize the model.
        step_func
            Function to apply to the autoencoder output at each step.
            inference_outputs, generative_outputs, losses, and a store variable are given to the function
        aggregation_func
            Function to aggregate the step results.
        indices
            Indices of cells in adata to use. If `None`, all cells are used.
        batch_size
            Minibatch size for data loading into model. Defaults to `scvi.settings.batch_size`.
        deterministic
            Makes model fully deterministic (e.g. no sampling in the bottleneck).
        """
        adata = self._validate_anndata(adata)
        data_loader = self._make_data_loader(adata=adata, indices=indices, batch_size=batch_size)

        store = []
        try:
            if deterministic:
                self.module.fully_deterministic = True
            for tensors in data_loader:
                loss_kwargs = {"kl_weight": 1}
                inference_outputs, generative_outputs, losses = self.module(tensors, loss_kwargs=loss_kwargs)
                step_func(inference_outputs, generative_outputs, losses, store)
        except Exception as e:
            self.module.fully_deterministic = False
            raise e
        finally:
            self.module.fully_deterministic = False

        return aggregation_func(store)

    @torch.inference_mode()
    def get_reconstruction_effect_of_each_split(
        self,
        adata: AnnData | None = None,
        add_to_counts: float = 1.0,
        aggregate_over_cells: bool = True,
        deterministic: bool = True,
        **kwargs,
    ) -> np.ndarray:
        r"""Return the effect of each split on the reconstructed per sample.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData.
            If `None`, defaults to the AnnData object used to initialize the model.
        add_to_counts
            Value to add to the counts before computing the logarithm.
        aggregate_over_cells
            Whether to aggregate the effect over cells and return a value per dimension.
        deterministic
            Makes model fully deterministic (e.g. no sampling in the bottleneck).
        kwargs
            Additional keyword arguments for the `iterate_on_ae_output` method.
        """

        def calculate_effect(inference_outputs, generative_outputs, losses, store):
            if self.module.split_aggregation == "logsumexp":
                log_mean_params = generative_outputs["original_params"]["mean"]  # n_samples x n_splits x n_genes
                log_mean_params = F.pad(
                    log_mean_params, (0, 0, 0, 1), value=np.log(add_to_counts)
                )  # n_samples x (n_splits + 1) x n_genes
                effect_share = -torch.log(1 - F.softmax(log_mean_params, dim=-2)[:, :-1, :]).sum(
                    dim=-1
                )  # n_samples x n_splits
            elif self.module.split_aggregation == "sum":
                effect_share = torch.abs(generative_outputs["original_params"]["mean"]).sum(
                    dim=-1
                )  # n_samples x n_splits
            else:
                raise NotImplementedError("Only logsumexp and sum aggregations are supported for now.")
            return store.append(effect_share.detach().cpu())

        def aggregate_effects(store):
            return torch.cat(store, dim=0).numpy(force=True)

        output = self.iterate_on_ae_output(
            adata=adata,
            step_func=calculate_effect,
            aggregation_func=aggregate_effects,
            deterministic=deterministic,
            **kwargs,
        )

        if aggregate_over_cells:
            output = output.sum(axis=0)

        return output

    # @torch.inference_mode()
    def get_max_effect_of_splits_within_distribution(
        self,
        adata: AnnData | None = None,
        add_to_counts: float = 1.0,
        deterministic: bool = True,
        **kwargs,
    ):
        r"""
        Return the max effect of each split on the reconstructed expression params for all genes.

        These values are empirical and inexact for de reasoning of dimensions.

        Parameters
        ----------
        adata
            AnnData object with equivalent structure to initial AnnData.
            If `None`, defaults to the AnnData object used to initialize the model.
        add_to_counts
            Value to add to the counts before computing the logarithm.
        deterministic
            Makes model fully deterministic (e.g. no sampling in the bottleneck).
        kwargs
            Additional keyword arguments for the `iterate_on_ae_output` method.

        Returns
        -------
        np.ndarray
            Max effect of each split on the reconstructed expression params for all genes


        -------
        Example usage
        -------

        import utils.plotting._interpretability        >>> effects = model.get_max_effect_of_splits_within_distribution(add_to_counts=0.1)
        >>> effect_data = (
        ...     pd.DataFrame(
        ...         effects,
        ...         columns=model.adata.var_names,
        ...         index=embed.var["title"],
        ...     )
        ...     .loc[embed.var.query("vanished == False").sort_values("order")["title"]]
        ...     .T
        ... )
        >>> plot_info = list(effect_data.to_dict(orient="series").items())
        >>> utils.plotting._interpretability._bar_plot_top_differential_vars(plot_info)
        >>> utils.plotting._interpretability._umap_of_relevant_genes(adata, embed, plot_info, dim_subset=["DR 1"])
        """

        def calculate_effect(inference_outputs, generative_outputs, losses, store):
            if self.module.split_aggregation == "logsumexp":
                log_mean_params = generative_outputs["original_params"]["mean"]  # n_samples x n_splits x n_genes
                log_mean_params = F.pad(
                    log_mean_params, (0, 0, 0, 1), value=np.log(add_to_counts)
                )  # n_samples x (n_splits + 1) x n_genes
                effect_share = -torch.log(1 - F.softmax(log_mean_params, dim=-2)[:, :-1, :])
            elif self.module.split_aggregation == "sum":
                effect_share = torch.abs(generative_outputs["original_params"]["mean"])
            else:
                raise NotImplementedError("Only logsumexp and sum aggregations are supported for now.")
            effect_share = effect_share.amax(dim=0).detach().cpu().numpy(force=True)
            if len(store) == 0:
                store.append(effect_share)
            else:
                store[0] = np.maximum(store[0], effect_share)

        def aggregate_effects(store):
            return store[0]

        output = self.iterate_on_ae_output(
            adata=adata,
            step_func=calculate_effect,
            aggregation_func=aggregate_effects,
            deterministic=deterministic,
            **kwargs,
        )

        return output



import math
import os
from typing import Literal
from uuid import uuid4

import merlin.io
import pandas as pd
import pyarrow.parquet as pq
from scvi.data import _constants

from drvi.scvi_tools_based.merlin_data._utils import read_first_row, transfer_type_from_pyarrow


class MerlinData:
    """
    Wrapper for merlin data object.

    Parameters
    ----------
    data_path : str
        Path to merlin data.
    train_key : str, optional
        Key for train data. Default is 'train'.
    validation_key : str, optional
        Key for validation data. Default is 'val'.
    test_key : str, optional
        Key for test data. Default is None.
    default_track : str, optional
        Default track to use for data. Default is 'train'. Possible values are 'train', 'val', 'test'.
    layer_key : str, optional
        Key for layer data. Default is None, which will be replaced with 'X'.
    sub_sample_frac : float
        The fraction of data to subsample. Default is 1.0.
    var_names_col_num : int
        The column number of the variable names in the var.parquet file. Default is 0.
    """

    def __init__(
        self,
        data_path: str,
        train_key: str = "train",
        validation_key: str = "val",
        test_key: str | None = "test",
        default_track: Literal["train", "val", "test"] = "train",
        layer_key: str | None = None,
        sub_sample_frac: float = 1.0,
        var_names_col_num=0,
    ) -> None:
        self.data_path = data_path
        self.train_key = train_key
        self.validation_key = validation_key
        self.test_key = test_key
        self.default_track = self.set_default_track(default_track)
        self.layer_key = "X" if layer_key is None else layer_key
        self.sub_sample_frac = sub_sample_frac

        self.var_names_col_num = var_names_col_num

        self.uns = {}
        self.schema, self.first_row = self.resolve_schema()

        assert self.schema.get(self.layer_key) is not None

    def get_default_track(self):
        return self.default_track

    def set_default_track(self, track):
        assert track in ["train", "val", "test"]
        self.default_track = track
        return track

    def resolve_schema(self):
        """
        Resolve the schema of the merlin data.

        Returns
        -------
        schema : dict
            The resolved schema of the merlin data.
        first_row : dict
            The first row of the merlin data.
        """
        single_parquet_filename = os.path.join(self.data_path, self.train_key, "part.0.parquet")
        first_row = read_first_row(single_parquet_filename)
        parquet_schema = pq.read_schema(single_parquet_filename)
        merlin_schema = transfer_type_from_pyarrow(parquet_schema, first_row=first_row)
        return merlin_schema, first_row

    def has_col(self, key):
        """
        Check if the merlin data has a specific column.

        Parameters
        ----------
        key : str
            The column key to check.

        Returns
        -------
        bool
            True if the column exists, False otherwise.
        """
        return self.schema.get(key) is not None

    def get_categorical_mapping(self, key):
        """
        Get the categorical mapping for a specific column.

        Parameters
        ----------
        key : str
            The column key to get the mapping for.

        Returns
        -------
        dict
            The categorical mapping for the column.
        """
        lookup_table = pd.read_parquet(os.path.join(self.data_path, "categorical_lookup", f"{key}.parquet"))
        lookup_table.columns = ["col_value"]
        mapping = lookup_table.to_dict()["col_value"]
        # correct mapping (int key to str) for compatbility with anndata summarization (attrdict)
        mapping = {str(k): v for k, v in mapping.items()}
        return mapping

    @property
    def n_vars(self):
        """
        Get the number of variables in the merlin data.

        Returns
        -------
        int
            The number of variables.
        """
        return self.schema.get(self.layer_key).value_count.max

    @property
    def var(self):
        """
        Get the variable metadata in the merlin data.

        Returns
        -------
        pd.DataFrame
            The variable metadata.
        """
        return pd.read_parquet(os.path.join(self.data_path, "var.parquet"))

    @property
    def var_names(self):
        """
        Get the variable names in the merlin data.

        Returns
        -------
        List[str]
            The variable names.
        """
        return self.var.iloc[:, self.var_names_col_num]

    @property
    def is_view(self):
        """For compatibility with anndata"""
        return False

    def _set_uuid(self, overwrite: bool = False):
        """
        Set the UUID for the merlin data.

        Parameters
        ----------
        overwrite : bool, optional
            Whether to overwrite the existing UUID. Default is False.
        """
        if _constants._SCVI_UUID_KEY not in self.uns or overwrite:
            self.uns[_constants._SCVI_UUID_KEY] = str(uuid4())

    def __repr__(self) -> str:
        return f"MerlinData object with schema: {self.schema.to_pandas()}"

    @staticmethod
    def _get_data_files(base_path: str, ds_key_disk: str, sub_sample_frac: float = 1.0):
        """
        Get the data files for a specific key.

        Parameters
        ----------
        base_path : str
            The base path of the data.
        ds_key_disk : str
            The key on disk to get the data files for.
        sub_sample_frac : float, optional
            The fraction of data to subsample. Default is 1.0.

        Returns
        -------
        str or List[str]
            The path(s) to the data file(s).
        """
        if sub_sample_frac == 1.0:
            # if no subsampling -> just return base path and merlin takes care of the rest
            return os.path.join(base_path, ds_key_disk)
        else:
            files = [file for file in os.listdir(os.path.join(base_path, ds_key_disk)) if file.endswith(".parquet")]
            files = [
                os.path.join(base_path, ds_key_disk, file) for file in sorted(files, key=lambda x: int(x.split(".")[1]))
            ]
            return files[: math.ceil(sub_sample_frac * len(files))]

    def get_dataset(self, split, columns, **dataset_kwargs):
        """
        Get the dataset for a specific split and columns.

        Parameters
        ----------
        split : str
            The split to get the dataset for.
        columns : List[str]
            The columns to include in the dataset.
        dataset_kwargs : dict, optional
            Additional keyword arguments to pass to the dataset.

        Returns
        -------
        merlin.io.Dataset
            The dataset object.
        """
        if split == "default":
            split = self.default_track
        sub_sample_frac = self.sub_sample_frac
        part_size = {
            "train": "100MB",
            "val": "100MB",
            "test": "300MB",
        }[split]
        ds_key_disk = {
            "train": self.train_key,
            "val": self.validation_key,
            "test": self.test_key,
        }[split]
        return merlin.io.Dataset(
            self._get_data_files(self.data_path, ds_key_disk=ds_key_disk, sub_sample_frac=sub_sample_frac),
            engine="parquet",
            part_size=part_size,
            schema=self.schema.select_by_name(columns),
            **dataset_kwargs,
        )



import logging
from collections import OrderedDict

import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.nn import functional as F


class FreezableEmbedding(nn.Embedding):
    def __init__(self, num_embeddings, embedding_dim, n_freeze_x=0, n_freeze_y=0, **kwargs):
        self._freeze_hook = None
        self.n_freeze_x = None
        self.n_freeze_y = None
        super().__init__(num_embeddings, embedding_dim, **kwargs)
        self.freeze(n_freeze_x, n_freeze_y)

    def freeze(self, n_freeze_x=0, n_freeze_y=0):
        self.n_freeze_x = n_freeze_x
        self.n_freeze_y = n_freeze_y

        if self._freeze_hook is not None:
            self._freeze_hook.remove()
            self._freeze_hook = None
        if self.n_freeze_x > 0 and self.n_freeze_y > 0:
            self._freeze_hook = self.weight.register_hook(self.partial_freeze_backward_hook)

    def partial_freeze_backward_hook(self, grad):
        with torch.no_grad():
            mask = F.pad(
                torch.zeros(self.n_freeze_x, self.n_freeze_y, device=grad.device),
                (0, self.embedding_dim - self.n_freeze_y, 0, self.num_embeddings - self.n_freeze_x),
                value=1.0,
            )
            return grad * mask

    def __repr__(self):
        if self._freeze_hook is None:
            return f"Emb({self.num_embeddings}, {self.embedding_dim})"
        else:
            return f"Emb({self.num_embeddings}, {self.embedding_dim} | freeze: {self.n_freeze_x}, {self.n_freeze_y})"


class MultiEmbedding(nn.Module):
    def __init__(
        self,
        n_embedding_list: list[int],
        embedding_dim_list: list[int],
        init_method="xavier_uniform",
        normalization=None,
        **kwargs,
    ):
        super().__init__()
        assert len(n_embedding_list) == len(embedding_dim_list)

        self.emb_list = nn.ParameterList(
            [
                FreezableEmbedding(n_embedding, embedding_dim, **kwargs)
                for n_embedding, embedding_dim in zip(n_embedding_list, embedding_dim_list, strict=False)
            ]
        )
        assert normalization in [None, "l2"]
        self.normalization = normalization
        self.reset_parameters(init_method)

    def reset_parameters(self, init_method):
        # TODO: research the correct distribution considering emb_side and layer_sizes
        for emb in self.emb_list:
            if init_method is None:
                pass
            elif callable(init_method):
                init_method(emb.weight)
            elif init_method == "xavier_uniform":
                nn.init.xavier_uniform_(emb.weight)
            elif init_method == "xavier_normal":
                nn.init.xavier_normal_(emb.weight)
            elif init_method == "uniform":
                nn.init.uniform_(emb.weight, -1.0, 1.0)
            elif init_method == "normal":
                nn.init.normal_(emb.weight)
            elif init_method == "zero":
                nn.init.zeros_(emb.weight)
            elif init_method == "one":
                nn.init._no_grad_fill_(emb.weight, 1.0)
            else:
                raise NotImplementedError()

    def forward(self, index_list: torch.Tensor) -> torch.Tensor:
        assert index_list.shape[-1] == len(self.emb_list)
        emb = torch.concat([emb(index_list[..., i]) for i, emb in enumerate(self.emb_list)], dim=-1)
        if self.normalization is None:
            return emb
        elif self.normalization == "l2":
            return F.normalize(emb, p=2, dim=1)

    @classmethod
    def from_pretrained(cls, feature_embedding_instance):
        raise NotImplementedError()

    def load_weights_from_trained_module(self, other, freeze_old=False):
        assert len(self.emb_list) >= len(other.emb_list)
        if len(self.emb_list) > len(other.emb_list):
            logging.warning(f"Extending feature embedding {other} to {self} with more feature categories.")
        else:
            logging.info(f"Extending feature embedding {other} to {self}")
        for self_emb, other_emb in zip(self.emb_list, other.emb_list, strict=False):
            assert self_emb.num_embeddings >= other_emb.num_embeddings
            with torch.no_grad():
                extension_size = (
                    0,
                    self_emb.embedding_dim - other_emb.embedding_dim,
                    0,
                    self_emb.num_embeddings - other_emb.num_embeddings,
                )
                transfer_mask = F.pad(
                    torch.zeros_like(other_emb.weight.data, device=other_emb.weight.data.device),
                    extension_size,
                    value=1.0,
                )
                extended_other_emb = F.pad(other_emb.weight.data, extension_size, value=0.0)
                self_emb.weight.data = extended_other_emb + transfer_mask * self_emb.weight.data
            if freeze_old:
                self_emb.freeze(other_emb.num_embeddings, other_emb.embedding_dim)

    def freeze_top_embs(self, n_freeze_list):
        for emb, n_freeze in zip(self.emb_list, n_freeze_list, strict=False):
            emb.freeze(n_freeze, emb.embedding_dim)

    @property
    def num_embeddings(self):
        return [emb.num_embeddings for emb in self.emb_list]

    @property
    def embedding_dim(self):
        return sum(emb.embedding_dim for emb in self.emb_list)

    def __repr__(self):
        repr_text = "cat(" + ", ".join(repr(emb) for emb in self.emb_list) + ")"
        if self.normalization:
            repr_text = f"{self.normalization}({repr_text})"
        return repr_text


class FeatureEmbedding(nn.Module):
    def __init__(self, vocab_list: list[list[str]], embedding_dims: list[int], **kwargs):
        super().__init__()
        assert len(vocab_list) == len(embedding_dims)
        self.device_container = nn.Parameter(torch.tensor([]))

        self.vocab_list = vocab_list
        n_vocab_list = [len(vocab) for vocab in self.vocab_list]
        self.multi_emb = self.define_embeddings(n_vocab_list, embedding_dims, **kwargs)

        self.__index_cache = {}
        self.__vocab_map_list_cache = None

    def reset_cache(self):
        self.__index_cache = {}
        self.__vocab_map_list_cache = None

    @property
    def vocab_map_list(self):
        if self.__vocab_map_list_cache is None:
            n_vocab_list = [len(vocab) for vocab in self.vocab_list]
            self.__vocab_map_list_cache = [
                dict(zip(vocab, range(n_vocab), strict=False))
                for vocab, n_vocab in zip(self.vocab_list, n_vocab_list, strict=False)
            ]
        return self.__vocab_map_list_cache

    @staticmethod
    def define_embeddings(n_embedding_list, embedding_dims, **kwargs):
        return MultiEmbedding(n_embedding_list, embedding_dims, **kwargs)

    def reset_parameters(self, init_method):
        self.multi_emb.reset_parameters(init_method)

    def _get_index_from_sentences(self, index_sentences: np.ndarray):
        assert index_sentences.shape[-1] == len(self.vocab_map_list)
        mapping_list = map(lambda mapping: np.vectorize(lambda key: mapping[key]), self.vocab_map_list)

        indices = torch.concat(
            [torch.from_numpy(mapping(index_sentences[..., [i]])) for i, mapping in enumerate(mapping_list)], dim=-1
        )
        return indices.to(self.device_container.device)

    def forward(self, index_sentences: np.ndarray, index_cache_key=None):
        if index_cache_key is not None and index_cache_key in self.__index_cache:
            return self.multi_emb(self.__index_cache[index_cache_key])
        indices = self._get_index_from_sentences(index_sentences)
        if index_cache_key is not None:
            self.__index_cache[index_cache_key] = indices
        return self.multi_emb(indices)

    def get_extra_state(self):
        return {"vocab_list": self.vocab_list}

    def set_extra_state(self, state):
        self.vocab_list = state["vocab_list"]
        self.__vocab_map_list_cache = None

    @classmethod
    def from_numpy_array(cls, sentences_array: np.ndarray, embedding_dims: np.ndarray | list, **kwargs):
        word_list = sentences_array.transpose().tolist()
        vocab_list = [list(OrderedDict.fromkeys(words)) for words in word_list]
        return cls(vocab_list, embedding_dims, **kwargs)

    @classmethod
    def from_pandas_dataframe(cls, sentences_df: pd.DataFrame, embedding_dims: pd.DataFrame | list, **kwargs):
        if isinstance(embedding_dims, pd.DataFrame):
            assert sentences_df.columns == embedding_dims.columns
            assert len(embedding_dims) == 1
            embedding_dims = embedding_dims.loc[0].to_list()
        return cls.from_numpy_array(sentences_df.values, embedding_dims, **kwargs)

    @classmethod
    def from_pretrained(cls, feature_embedding_instance):
        raise NotImplementedError()

    def load_weights_from_trained_module(self, other, freeze_old=False):
        assert isinstance(other, self.__class__)
        assert len(self.vocab_list) >= len(other.vocab_list)

        for self_vocab, other_vocab in zip(self.vocab_list, other.vocab_list, strict=False):
            assert len(self_vocab) >= len(other_vocab)
            for self_word, other_word in zip(self_vocab, other_vocab, strict=False):
                assert self_word == other_word

        self.multi_emb.load_weights_from_trained_module(other.multi_emb, freeze_old=freeze_old)

    @property
    def embedding_dim(self):
        return self.multi_emb.embedding_dim

    def __repr__(self):
        return f"str2index -> {repr(self.multi_emb)}"



from ._differential_vars import (
    calculate_differential_vars,
    combine_differential_effects,
    find_differential_effects,
    get_split_effects,
    iterate_on_top_differential_vars,
)
from ._latent_traverse import traverse_latent

__all__ = [
    "traverse_latent",
    "find_differential_effects",
    "combine_differential_effects",
    "calculate_differential_vars",
    "iterate_on_top_differential_vars",
    "get_split_effects",
]



import itertools
from collections.abc import Sequence

import numpy as np
import pandas as pd
import scanpy as sc
from anndata import AnnData
from matplotlib import pyplot as plt

from drvi.utils.plotting import cmap
from drvi.utils.tools import iterate_on_top_differential_vars
from drvi.utils.tools.interpretability._latent_traverse import get_dimensions_of_traverse_data


def make_heatmap_groups(ordered_list):
    n_groups, group_names = zip(
        *[(len(list(group)), key) for (key, group) in itertools.groupby(ordered_list)], strict=False
    )
    group_positions = [0] + list(itertools.accumulate(n_groups))
    group_positions = list(zip(group_positions[:-1], [c - 1 for c in group_positions[1:]], strict=False))
    return group_positions, group_names


def differential_vars_heatmap(
    traverse_adata: AnnData,
    key: str,
    title_col: str = "title",
    score_threshold: float = 0.0,
    remove_vanished: bool = True,
    remove_unaffected: bool = False,
    figsize: tuple[int, int] | None = None,
    show: bool = True,
    **kwargs,
):
    """
    Generate a heatmap of differential variables based on traverse data.

    Parameters
    ----------
    - traverse_adata (AnnData): Annotated data object containing traverse data.
    - key (str): Key used to access traverse effect data in `traverse_adata.varm`.
    - title_col (str): Column name in `traverse_adata.obs` to use as dimension labels.
    - score_threshold (float): Threshold value for filtering variables based on the score.
    - remove_vanished (bool): Whether to remove variables that have vanished.
    - remove_unaffected (bool): Whether to remove variables that have no effect.
    - figsize (Optional[Tuple[int, int]]): Size of the figure (width, height).
    - show (bool): Whether to show the heatmap.
    - **kwargs: Additional keyword arguments to be passed to `sc.pl.heatmap`.

    Returns
    -------
    - None if show is True, otherwise the plot.
    """
    n_latent, n_steps, n_samples, n_vars = get_dimensions_of_traverse_data(traverse_adata)

    max_effect_index_in_positive_direction_for_each_gene = np.abs(
        traverse_adata.varm[f"{key}_traverse_effect_pos"]
    ).values.argmax(axis=1)
    max_effect_in_positive_direction_for_each_gene = np.abs(
        traverse_adata.varm[f"{key}_traverse_effect_pos"].values[
            np.arange(traverse_adata.n_vars), max_effect_index_in_positive_direction_for_each_gene
        ]
    )
    max_effect_index_in_negative_direction_for_each_gene = np.abs(
        traverse_adata.varm[f"{key}_traverse_effect_neg"]
    ).values.argmax(axis=1)
    max_effect_in_negative_direction_for_each_gene = np.abs(
        traverse_adata.varm[f"{key}_traverse_effect_neg"].values[
            np.arange(traverse_adata.n_vars), max_effect_index_in_negative_direction_for_each_gene
        ]
    )
    traverse_adata.var["max_effect"] = np.maximum(
        max_effect_in_positive_direction_for_each_gene, max_effect_in_negative_direction_for_each_gene
    )
    for col in ["dim_id", "order", title_col]:
        title_mapping = dict(zip(traverse_adata.obs["dim_id"].values, traverse_adata.obs[col].values, strict=False))
        traverse_adata.var[f"max_effect_dim_{col}"] = np.where(
            traverse_adata.var["max_effect"] < score_threshold,
            float("nan") if np.isreal(traverse_adata.obs[col].values[0]) else "NONE",
            np.where(
                max_effect_in_positive_direction_for_each_gene > max_effect_in_negative_direction_for_each_gene,
                pd.Series(max_effect_index_in_positive_direction_for_each_gene).map(title_mapping),
                pd.Series(max_effect_index_in_negative_direction_for_each_gene).map(title_mapping),
            ),
        )
        traverse_adata.var[f"max_effect_dim_{col}_plus"] = np.where(
            traverse_adata.var["max_effect"] < score_threshold,
            "NONE",
            np.where(
                max_effect_in_positive_direction_for_each_gene > max_effect_in_negative_direction_for_each_gene,
                pd.Series(max_effect_index_in_positive_direction_for_each_gene).map(title_mapping).astype(str) + " +",
                pd.Series(max_effect_index_in_negative_direction_for_each_gene).map(title_mapping).astype(str) + " -",
            ),
        )

    plot_adata = AnnData(
        traverse_adata.uns[f"{key}_traverse_effect_stepwise"].reshape(n_latent * n_steps, n_vars),
        var=traverse_adata.var,
        obs=pd.DataFrame(
            {
                "dim_id": np.repeat(np.arange(n_latent), n_steps),
                "step_id": np.tile(np.arange(n_steps), n_latent),
            }
        ),
    )
    for col in ["dim_id", "order", title_col, "vanished"]:
        title_mapping = dict(zip(traverse_adata.obs["dim_id"].values, traverse_adata.obs[col].values, strict=False))
        plot_adata.obs[col] = plot_adata.obs["dim_id"].map(title_mapping)

    if remove_vanished:
        plot_adata = plot_adata[~plot_adata.obs["vanished"]].copy()
    if remove_unaffected:
        plot_adata = plot_adata[:, plot_adata.var["max_effect"] > score_threshold].copy()
    plot_adata = plot_adata[
        :, plot_adata.var.sort_values(["max_effect_dim_order", "max_effect_dim_order_plus", "max_effect"]).index
    ].copy()
    plot_adata = plot_adata[plot_adata.obs.sort_values(["order"]).index].copy()

    if figsize is None:
        figsize = (20, plot_adata.obs["dim_id"].nunique() / 4)

    vmin = min(
        -1,
        min(
            traverse_adata.varm[f"{key}_traverse_effect_pos"].values.min(),
            traverse_adata.varm[f"{key}_traverse_effect_neg"].values.min(),
        ),
    )
    vmax = max(
        +1,
        max(
            traverse_adata.varm[f"{key}_traverse_effect_pos"].values.max(),
            traverse_adata.varm[f"{key}_traverse_effect_neg"].values.max(),
        ),
    )
    var_group_positions, var_group_labels = make_heatmap_groups(plot_adata.var[f"max_effect_dim_{title_col}_plus"])
    kwargs = {
        **dict(  # noqa: C408
            vcenter=0,
            vmin=vmin,
            vmax=vmax,
            cmap=cmap.saturated_red_blue_cmap,
            var_group_positions=var_group_positions,
            var_group_labels=var_group_labels,
            var_group_rotation=90,
        ),
        **kwargs,
    }

    return sc.pl.heatmap(
        plot_adata,
        plot_adata.var.index,
        groupby=title_col,
        layer=None,
        figsize=figsize,
        dendrogram=False,
        show=show,
        **kwargs,
    )


def _bar_plot_top_differential_vars(
    plot_info: Sequence[tuple[str, pd.Series]],
    dim_subset: Sequence[str] | None = None,
    n_top_genes: int = 10,
    ncols: int = 5,
    show: bool = True,
):
    """
    Plot the top differential variables in a bar plot.

    Parameters
    ----------
        plot_info (Sequence[Tuple[str, pd.Series]]): Information about the top differential variables.
        dim_subset (Sequence[sre]): List of dimensions to plot in the bar plot. If not specified all dimensions are plotted.
        n_top_genes (int, optional): Number of top genes to plot. Defaults to 10.
        ncols (int, optional): Number of columns in the plot grid. Defaults to 5.
        show (bool, optional): Whether to display the plot. If False, the plot will be returned as a Figure object. Defaults to True.

    Returns
    -------
        None if show is True, otherwise the figure.
    """
    if dim_subset is not None:
        plot_info = dict(plot_info)
        plot_info = [(dim_id, plot_info[dim_id]) for dim_id in dim_subset]

    n_row = int(np.ceil(len(plot_info) / ncols))
    fig, axes = plt.subplots(n_row, ncols, figsize=(3 * ncols, int(1 + 0.2 * n_top_genes) * n_row))

    for ax, info in zip(axes.flatten(), plot_info, strict=False):
        dim_title = info[0]

        top_indices = info[1].sort_values(ascending=False)[:n_top_genes]
        genes = top_indices.index
        values = top_indices.values

        # Create a horizontal bar plot
        ax.barh(genes, values, color="skyblue")
        ax.set_xlabel("Gene Score")
        ax.set_title(dim_title)
        ax.invert_yaxis()
        ax.grid(False)

    for ax in axes.flatten()[len(plot_info) :]:
        fig.delaxes(ax)

    plt.tight_layout()
    if show:
        plt.show()
    else:
        return fig


def show_top_differential_vars(
    traverse_adata: AnnData,
    key: str,
    title_col: str = "title",
    order_col: str = "order",
    dim_subset: Sequence[str] | None = None,
    gene_symbols: str | None = None,
    score_threshold: float = 0.0,
    n_top_genes: int = 10,
    ncols: int = 5,
    show: bool = True,
):
    """
    Show top differential variables in a bar plot.

    Parameters
    ----------
        traverse_adata (AnnData): Annotated data object containing the variables to be plotted.
        key (str): Key to access the traverse effect variables in `traverse_adata.varm`.
        title_col (str, optional): Column name in `traverse_adata.obs` that contains the titles for each dimension. Defaults to 'title'.
        order_col (str, optional): Column name in `traverse_adata.obs` that specifies the order of dimensions. Defaults to 'order'.  Ignored if `dim_subset` is provided.
        dim_subset (Sequence[sre]): List of dimensions to plot in the bar plot. If not specified all dimensions are plotted.
        gene_symbols (str, optional): Column name in `traverse_adata.var` that contains gene symbols. If provided, gene symbols will be used in the plot instead of gene indices. Defaults to None.
        score_threshold (float, optional): Threshold value for gene scores. Only genes with scores above this threshold will be plotted. Defaults to 0.
        n_top_genes (int, optional): Number of top genes to plot. Defaults to 10.
        ncols (int, optional): Number of columns in the plot grid. Defaults to 5.
        show (bool, optional): Whether to display the plot. If False, the plot will be returned as a Figure object. Defaults to True.

    Returns
    -------
        None if show is True, otherwise the figure.
    """
    plot_info = iterate_on_top_differential_vars(
        traverse_adata, key, title_col, order_col, gene_symbols, score_threshold
    )

    return _bar_plot_top_differential_vars(plot_info, dim_subset, n_top_genes, ncols, show)


def show_differential_vars_scatter_plot(
    traverse_adata: AnnData,
    key_x: str,
    key_y: str,
    key_combined: str,
    title_col: str = "title",
    order_col: str = "order",
    gene_symbols: str | None = None,
    score_threshold: float = 0.0,
    dim_subset: Sequence[str] | None = None,
    ncols: int = 3,
    show: bool = True,
    **kwargs,
):
    """
    Show a scatter plot of differential variables conidering multiple criteria.

    Parameters
    ----------
    - traverse_adata (AnnData): Annotated data object containing the variables to be plotted.
    - key_x (str): Key to access the first variable in `traverse_adata.varm`.
    - key_y (str): Key to access the second variable in `traverse_adata.varm`.
    - key_combined (str): Key to access the combined variable in `traverse_adata.varm`.
    - title_col (str, optional): Column name in `traverse_adata.obs` that contains the titles for each dimension. Defaults to 'title'.
    - order_col (str, optional): Column name in `traverse_adata.obs` that specifies the order of dimensions. Defaults to 'order'.  Ignored if `dim_subset` is provided.
    - gene_symbols (str, optional): Column name in `traverse_adata.var` that contains gene symbols. If provided, gene symbols will be used in the plot instead of gene indices. Defaults to None.
    - score_threshold (float, optional): Threshold value for gene scores. Only genes with scores above this threshold will be plotted. Defaults to 0.
    - dim_subset (Optional[Sequence[str]], optional): Subset of dimensions to plot. If None, all dimensions will be plotted. Defaults to None.
    - ncols (int, optional): Number of columns in the plot grid. Defaults to 3.
    - show (bool, optional): Whether to display the plot. If False, the plot will be returned as a Figure object. Defaults to True.
    - **kwargs: Additional keyword arguments to be passed to the scatter plot.

    Returns
    -------
    - None if show is True, otherwise the figure.
    """
    plot_info = {}
    for key in [key_x, key_y, key_combined]:
        plot_info[key] = iterate_on_top_differential_vars(
            traverse_adata, key, title_col, order_col, gene_symbols, score_threshold
        )

    if dim_subset is None:
        dim_ids = [dim_id for dim_id, _ in plot_info[key_combined]]
    else:
        dim_ids = [dim_id for dim_id, _ in plot_info[key_combined] if dim_id in dim_subset]
    for key in [key_x, key_y, key_combined]:
        plot_info[key] = dict(plot_info[key])

    n_plots = len(dim_ids)
    n_row = int(np.ceil(n_plots / ncols))
    fig, axes = plt.subplots(n_row, ncols, figsize=(5 * ncols, 4 * n_row), sharex=False, sharey=False)

    for ax, dim_id in zip(axes.flatten(), dim_ids, strict=False):
        df = (
            pd.concat(
                {
                    key_x: plot_info[key_x][dim_id],
                    key_y: plot_info[key_y][dim_id],
                    key_combined: plot_info[key_combined][dim_id],
                },
                axis=1,
            )
            .dropna()
            .sort_values(key_combined, ascending=False)
        )

        scatter = ax.scatter(df[key_x], df[key_y], c=df[key_combined], cmap="Reds", **kwargs)

        cbar = fig.colorbar(scatter, ax=ax)
        cbar.set_label(key_combined)

        top_20 = df.nlargest(20, key_combined)
        for idx, row in top_20.iterrows():
            ax.text(row[key_x], row[key_y], str(idx), fontsize=8, ha="left", va="bottom")

        ax.set_title(dim_id)
        ax.set_xlabel(key_x)
        ax.set_ylabel(key_y)

    for ax in axes.flatten()[n_plots:]:
        fig.delaxes(ax)

    plt.tight_layout()
    if show:
        plt.show()
    else:
        return fig


def _umap_of_relevant_genes(
    adata: AnnData,
    embed: AnnData,
    plot_info: Sequence[tuple[str, pd.Series]],
    layer: str | None = None,
    title_col: str = "title",
    gene_symbols: str | None = None,
    dim_subset: Sequence[str] | None = None,
    n_top_genes: int = 10,
    max_cells_to_plot: int | None = None,
    **kwargs,
):
    if max_cells_to_plot is not None and adata.n_obs > max_cells_to_plot:
        adata = sc.pp.subsample(adata, n_obs=max_cells_to_plot, copy=True)

    if dim_subset is not None:
        plot_info = dict(plot_info)
        plot_info = [(dim_id, plot_info[dim_id]) for dim_id in dim_subset]

    adata.obsm["X_umap_method"] = embed[adata.obs.index].obsm["X_umap"]
    for dim_title, gene_scores in plot_info:
        print(dim_title)
        relevant_genes = gene_scores.sort_values(ascending=False).index.to_list()

        if dim_title[-1] == "+":
            _cmap = cmap.saturated_sky_cmap
            real_dim_title = dim_title[:-1]
        elif dim_title[-1] == "-":
            _cmap = cmap.saturated_sky_cmap.reversed()
            real_dim_title = dim_title[:-1]
        else:
            _cmap = cmap.saturated_red_blue_cmap
            real_dim_title = dim_title

        adata.obs[dim_title] = list(embed[adata.obs.index, embed.var[title_col] == real_dim_title].X[:, 0])
        ax = sc.pl.embedding(
            adata, "X_umap_method", color=[dim_title], cmap=_cmap, vcenter=0, show=False, frameon=False, **kwargs
        )
        ax.text(0.92, 0.05, ax.get_title(), size=15, ha="left", color="black", rotation=90, transform=ax.transAxes)
        ax.set_title("")
        plt.show()

        axes = sc.pl.embedding(
            adata,
            "X_umap_method",
            layer=layer,
            color=relevant_genes[:n_top_genes],
            cmap=cmap.saturated_just_sky_cmap,
            gene_symbols=gene_symbols,
            show=False,
            frameon=False,
            **kwargs,
        )
        if n_top_genes == 1 or len(relevant_genes) == 1:
            axes = [axes]
        for ax in axes:
            ax.text(0.92, 0.05, ax.get_title(), size=15, ha="left", color="black", rotation=90, transform=ax.transAxes)
            ax.set_title("")
        plt.show()


def plot_relevant_genes_on_umap(
    adata: AnnData,
    embed: AnnData,
    traverse_adata: AnnData,
    traverse_adata_key: str,
    layer: str | None = None,
    title_col: str = "title",
    order_col: str = "order",
    gene_symbols: str | None = None,
    score_threshold: float = 0.0,
    dim_subset: Sequence[str] = None,
    n_top_genes: int = 10,
    max_cells_to_plot: int | None = None,
    **kwargs,
):
    plot_info = iterate_on_top_differential_vars(
        traverse_adata, traverse_adata_key, title_col, order_col, gene_symbols, score_threshold
    )

    return _umap_of_relevant_genes(
        adata,
        embed,
        plot_info,
        layer,
        title_col,
        gene_symbols,
        dim_subset,
        n_top_genes,
        max_cells_to_plot,
        **kwargs,
    )



import numpy as np
import pandas as pd
from scipy import stats
from sklearn.feature_selection import mutual_info_classif


def check_discrete_metric_input(gt_cat_series=None, gt_one_hot=None):
    if gt_cat_series is not None and gt_one_hot is not None:
        raise ValueError("Only one of gt_cat_series or gt_one_hot should be provided.")
    if gt_cat_series is None and gt_one_hot is None:
        raise ValueError("Either gt_cat_series or gt_one_hot must be provided.")


def get_one_hot_encoding(gt_cat_series):
    return np.eye(len(gt_cat_series.cat.categories))[gt_cat_series.cat.codes]


def _nn_alignment_score_per_dim(var_continues, gt_01):
    order = var_continues.argsort()
    gt_01 = gt_01[order]
    alignment = np.clip(
        (
            np.sum(gt_01[:-1, :] * gt_01[1:, :], axis=0) / (np.sum(gt_01, axis=0) - 1)
        )  # fraction of cells of this type that are next to a cell of the same type
        - (np.sum(gt_01, axis=0) / gt_01.shape[0]),  # cancel random neighbors when GT (ground-truth) is frequent
        0,
        None,
    ) / (1 - (np.sum(gt_01, axis=0) / gt_01.shape[0]))
    return alignment


def nn_alignment_score(all_vars_continues, gt_cat_series=None, gt_one_hot=None):
    check_discrete_metric_input(gt_cat_series, gt_one_hot)
    gt_01 = get_one_hot_encoding(gt_cat_series) if gt_cat_series is not None else gt_one_hot

    n_vars = all_vars_continues.shape[1]
    result = np.zeros([n_vars, gt_01.shape[1]])
    for i in range(n_vars):
        result[i, :] = _nn_alignment_score_per_dim(all_vars_continues[:, i], gt_01)
    return result


def _local_mutual_info_score_per_binary_gt(all_vars_continues, gt_binary):
    mi_score = mutual_info_classif(all_vars_continues, gt_binary, n_jobs=-1)
    gt_prob = np.sum(gt_binary == 1) / gt_binary.shape[0]
    gt_entropy = stats.entropy([gt_prob, 1 - gt_prob])
    return mi_score / gt_entropy


def local_mutual_info_score(all_vars_continues, gt_cat_series=None, gt_one_hot=None):
    check_discrete_metric_input(gt_cat_series, gt_one_hot)
    gt_01 = get_one_hot_encoding(gt_cat_series) if gt_cat_series is not None else gt_one_hot

    n_vars = all_vars_continues.shape[1]
    result = np.zeros([n_vars, gt_01.shape[1]])
    for j in range(gt_01.shape[1]):
        result[:, j] = _local_mutual_info_score_per_binary_gt(all_vars_continues, gt_01[:, j])
    return result


def spearman_correlataion_score(all_vars_continues, gt_cat_series=None, gt_one_hot=None):
    check_discrete_metric_input(gt_cat_series, gt_one_hot)
    gt_01 = get_one_hot_encoding(gt_cat_series) if gt_cat_series is not None else gt_one_hot

    n_vars = all_vars_continues.shape[1]
    result = np.abs(stats.spearmanr(all_vars_continues, gt_01).statistic[:n_vars, n_vars:])
    return result


def global_dim_mutual_info_score(all_vars_continues, gt_cat_series):
    # This metric is not used in any analysis, but is provided for completeness.
    mi_score = mutual_info_classif(all_vars_continues, gt_cat_series)
    gt_entropy = stats.entropy(pd.Series(gt_cat_series).value_counts(normalize=True, sort=False))
    return mi_score / gt_entropy



from scvi.data.fields import CategoricalObsField, LayerField, NumericalJointObsField

from drvi.scvi_tools_based.data.fields import FixedCategoricalJointObsField
from drvi.scvi_tools_based.merlin_data._data import MerlinData


class MerlinLayerField(LayerField):
    def validate_field(self, adata: MerlinData) -> None:
        return

    def register_field(self, adata: MerlinData) -> dict:
        return {
            # self.N_OBS_KEY: adata.n_obs, ?
            self.N_VARS_KEY: adata.n_vars,
            # self.COLUMN_NAMES_KEY: np.asarray(adata.var_names), ?
        }

    def get_summary_stats(self, state_registry: dict) -> dict:
        summary_stats = {self.count_stat_key: state_registry[self.N_VARS_KEY]}
        # if self.registry_key == REGISTRY_KEYS.X_KEY:
        #     summary_stats[self.N_CELLS_KEY] = state_registry[self.N_OBS_KEY]
        return summary_stats

    def transfer_field(self, state_registry: dict, adata_target: MerlinData, **kwargs) -> dict:
        return super().transfer_field(state_registry, adata_target, **kwargs)


class MerlinCategoricalObsField(CategoricalObsField):
    def _setup_default_attr(self, adata: MerlinData) -> None:
        return

    def validate_field(self, adata: MerlinData) -> None:
        if self.is_default:
            return
        merlin_data = adata
        assert merlin_data.has_col(self.attr_key)

    def register_field(self, adata: MerlinData) -> dict:
        if self.is_default:
            return {
                self.CATEGORICAL_MAPPING_KEY: {},
            }
        merlin_data = adata
        categorical_mapping = merlin_data.get_categorical_mapping(self.attr_key)
        return {
            self.CATEGORICAL_MAPPING_KEY: categorical_mapping,
            self.ORIGINAL_ATTR_KEY: self._original_attr_key,
        }

    def get_summary_stats(self, state_registry: dict) -> dict:
        if self.is_default:
            return {}
        categorical_mapping = state_registry[self.CATEGORICAL_MAPPING_KEY]
        n_categories = len(categorical_mapping)
        return {self.count_stat_key: n_categories}

    def transfer_field(
        self,
        state_registry: dict,
        adata_target: MerlinData,
        extend_categories: bool = False,
        **kwargs,
    ) -> dict:
        """Transfer field from registry to target AnnData."""
        if self.is_default:
            self._setup_default_attr(adata_target)

        self.validate_field(adata_target)

        mapping = state_registry[self.CATEGORICAL_MAPPING_KEY].copy()

        assert not extend_categories

        return {
            self.CATEGORICAL_MAPPING_KEY: mapping,
            self.ORIGINAL_ATTR_KEY: self._original_attr_key,
        }


class MerlinCategoricalJointObsField(FixedCategoricalJointObsField):
    def validate_field(self, adata: MerlinData) -> None:
        merlin_data = adata
        for key in self.attr_keys:
            assert merlin_data.has_col(key)

    def register_field(self, adata: MerlinData) -> dict:
        merlin_data = adata
        categories = {}
        for key in self.attr_keys:
            mapping = merlin_data.get_categorical_mapping(key)
            categories[key] = mapping

        store_cats = categories

        mappings_dict = self._default_mappings_dict()
        mappings_dict[self.MAPPINGS_KEY] = store_cats
        mappings_dict[self.FIELD_KEYS_KEY] = self.attr_keys
        for k in self.attr_keys:
            mappings_dict[self.N_CATS_PER_KEY].append(len(store_cats[k]))
        return mappings_dict

    def get_summary_stats(self, state_registry: dict) -> dict:
        return super().get_summary_stats(state_registry)

    def transfer_field(
        self,
        state_registry: dict,
        adata_target: MerlinData,
        extend_categories: bool = False,
        **kwargs,
    ) -> dict:
        """Transfer the field."""
        if self.is_empty:
            return

        source_cat_dict = state_registry[self.MAPPINGS_KEY].copy()
        assert not extend_categories

        return source_cat_dict


class MerlinNumericalJointObsField(NumericalJointObsField):
    def validate_field(self, adata: MerlinData) -> None:
        merlin_data = adata
        for key in self.attr_keys:
            assert merlin_data.has_col(key)

    def register_field(self, adata: MerlinData) -> dict:
        if self.attr_keys:
            raise NotImplementedError()
        return {}

    def get_summary_stats(self, state_registry: dict) -> dict:
        if self.attr_keys:
            raise NotImplementedError()
        return {}

    def transfer_field(
        self,
        state_registry: dict,
        adata_target: MerlinData,
        **kwargs,
    ) -> dict:
        """Transfer the field."""
        super().transfer_field(state_registry, adata_target, **kwargs)
        return self.register_field(adata_target)


MerlinDataField = (
    MerlinLayerField | MerlinCategoricalObsField | MerlinCategoricalJointObsField | MerlinNumericalJointObsField
)



import re
from collections import namedtuple

import anndata as ad
import numpy as np
import pandas as pd

FeatureInfo = namedtuple("FeatureInfo", ["name", "dim", "keywords"])


class FeatureInfoList:
    def __init__(self, feature_info_str_list: list[str], axis="var", total_dim=None, default_dim=None):
        assert axis in ["var", "obs"]
        assert total_dim is None or default_dim is None
        self.feature_info_list = list(self.parse(feature_info_str_list))
        self.axis = axis
        if any(fi.dim is None for fi in self.feature_info_list):
            if total_dim is None and default_dim is None:
                raise ValueError(f"missing dim in {feature_info_str_list}\nPlease provide `total_dim` or `default_dim`")
            if total_dim is not None:
                self._fill_with_total_dim(total_dim)
            if default_dim is not None:
                self._fill_with_default_dim(default_dim)

    @staticmethod
    def parse(feature_info_list):
        pattern = re.compile(r"\A(?P<name>\w+)(@(?P<dim>\d*))?(?P<kw>(!\w+)+)?\Z")

        for feature_info in feature_info_list:
            match = pattern.match(feature_info)
            name, dim, kw = match.group("name"), match.group("dim"), match.group("kw")
            dim = None if dim is None else int(dim)
            kw = () if kw is None else tuple(kw[1:].split("!"))
            yield FeatureInfo(name, dim, kw)

    def _fill_with_default_dim(self, default_dim):
        for i in range(len(self.feature_info_list)):
            if self.feature_info_list[i].dim is None:
                self.feature_info_list[i] = self.feature_info_list[i]._replace(dim=default_dim)

    def _fill_with_total_dim(self, total_dim):
        n_none_dims = sum([(fi.dim is None) + 0.0 for fi in self.feature_info_list])
        if n_none_dims > 0:
            remaining_dims = total_dim - sum([fi.dim for fi in self.feature_info_list if fi.dim is not None])
            assert remaining_dims >= 0 and remaining_dims % n_none_dims == 0
            fill_value = int(remaining_dims // n_none_dims)
            self._fill_with_default_dim(fill_value)
        assert sum([fi.dim for fi in self.feature_info_list]) == total_dim

    @property
    def names(self):
        return [fi.name for fi in self.feature_info_list]

    @property
    def dims(self):
        return [fi.dim for fi in self.feature_info_list]

    @property
    def keywords_list(self):
        return [fi.keywords for fi in self.feature_info_list]

    def get_possible_values_array(self, data: dict[str, ad.AnnData]):
        keys = self.names
        if keys is None or len(keys) == 0:
            return np.array([])
        possible_vals = []
        if isinstance(data, ad.AnnData):
            data = {"dummy": data}
        for _, adada in data.items():
            if self.axis == "obs":
                df = adada.obs
            elif self.axis == "var":
                df = adada.var
            else:
                raise ValueError("axis should be either `obs` or `var`")
            possible_vals.append(df[list(keys)].drop_duplicates(subset=keys))
        return pd.concat(possible_vals).drop_duplicates(subset=keys).to_numpy()

    def __len__(self):
        return len(self.feature_info_list)

    def __iter__(self):
        return self.feature_info_list.__iter__()

    def __repr__(self):
        return repr(self.feature_info_list)



from . import _cmap as cmap
from ._interpretability import (
    differential_vars_heatmap,
    make_heatmap_groups,
    plot_relevant_genes_on_umap,
    show_differential_vars_scatter_plot,
    show_top_differential_vars,
)
from ._latent import (
    make_balanced_subsample,
    plot_latent_dimension_stats,
    plot_latent_dims_in_heatmap,
    plot_latent_dims_in_umap,
)

__all__ = [
    "make_balanced_subsample",
    "plot_latent_dimension_stats",
    "plot_latent_dims_in_umap",
    "plot_latent_dims_in_heatmap",
    "make_heatmap_groups",
    "differential_vars_heatmap",
    "show_top_differential_vars",
    "show_differential_vars_scatter_plot",
    "plot_relevant_genes_on_umap",
    "cmap",
]



from ._aggregation import latent_matching_score, most_similar_averaging_score, most_similar_gap_score
from ._benchmark import DiscreteDisentanglementBenchmark
from ._pairwise import (
    global_dim_mutual_info_score,
    local_mutual_info_score,
    nn_alignment_score,
    spearman_correlataion_score,
)

__all__ = [
    "nn_alignment_score",
    "local_mutual_info_score",
    "global_dim_mutual_info_score",
    "spearman_correlataion_score",
    "most_similar_averaging_score",
    "latent_matching_score",
    "most_similar_gap_score",
    "DiscreteDisentanglementBenchmark",
]



import numpy as np
from scipy.sparse import csr_matrix
from scipy.sparse.csgraph import min_weight_full_bipartite_matching


def most_similar_averaging_score(result_matrix):
    return np.mean(np.max(result_matrix, axis=0))


def latent_matching_score(result_matrix):
    row_ind, col_ind = min_weight_full_bipartite_matching(csr_matrix(-result_matrix - 1e-10))
    return result_matrix[row_ind, col_ind].sum() / result_matrix.shape[1]


def most_similar_gap_score(result_matrix):
    sorted_values = np.sort(result_matrix, axis=0)[::-1, :]
    return np.mean(sorted_values[0, :] - sorted_values[1, :])



import logging

from rich.console import Console
from rich.logging import RichHandler

from . import merlin_data, model, module

logger = logging.getLogger(__name__)
# set the logging level
logger.setLevel(logging.INFO)

# nice logging outputs
console = Console(force_terminal=True)
if console.is_jupyter is True:
    console.is_jupyter = False
ch = RichHandler(show_path=False, console=console, show_time=False)
formatter = logging.Formatter("DRVI: %(message)s")
ch.setFormatter(formatter)
logger.addHandler(ch)

# this prevents double outputs
logger.propagate = False

__all__ = ["model", "module", "merlin_data"]



import numpy as np
import pandas as pd
import scvi
from anndata import AnnData
from scipy import sparse

from drvi.model import DRVI


def iterate_dimensions(
    # n_latent: int,
    latent_dims: np.ndarray,
    latent_min: np.ndarray,
    latent_max: np.ndarray,
    n_steps: int = 10 * 2,
    n_samples: int = 100,
):
    assert n_steps % 2 == 0
    assert np.all(latent_min <= 0) & np.all(latent_max >= 0)

    dim_ids = (
        latent_dims.reshape(-1, 1, 1)
        * np.ones(n_steps).astype(int).reshape(1, -1, 1)
        * np.ones(n_samples).astype(int).reshape(1, 1, -1)
    ).reshape(-1)  # n_latent * n_steps * n_samples
    sample_ids = (
        np.ones(len(latent_dims)).astype(int).reshape(-1, 1, 1)
        * np.ones(n_steps).astype(int).reshape(1, -1, 1)
        * np.arange(n_samples).astype(int).reshape(1, 1, -1)
    ).reshape(-1)  # n_latent * n_steps * n_samples
    step_ids = (
        np.ones(len(latent_dims)).astype(int).reshape(-1, 1, 1)
        * np.arange(n_steps).astype(int).reshape(1, -1, 1)
        * np.ones(n_samples).astype(int).reshape(1, 1, -1)
    ).reshape(-1)  # n_latent * n_steps * n_samples
    span_values = (
        np.concatenate(
            [
                np.linspace(latent_min, latent_min * 0.0, num=int(n_steps / 2)),
                np.linspace(latent_max * 0.0, latent_max, num=int(n_steps / 2)),
            ],
            axis=0,
        ).T.reshape(-1, 1)
        * np.ones(n_samples).reshape(1, -1)
    ).reshape(-1)  # n_latent * n_steps * n_samples

    span_vectors = sparse.coo_matrix((span_values, (np.arange(len(dim_ids)), dim_ids)), dtype=np.float32)
    span_vectors = span_vectors.tocsr()  # n_latent * n_steps * n_samples x n_latent

    span_adata = AnnData(
        X=span_vectors,
        obs=pd.DataFrame(
            {
                "original_order": np.arange(span_vectors.shape[0]),
                "dim_id": dim_ids,
                "sample_id": sample_ids,
                "step_id": step_ids,
                "span_value": span_values,
            }
        ),
    )

    return span_adata


def make_traverse_adata(
    model: DRVI,
    embed: AnnData,
    n_steps: int = 10 * 2,
    n_samples: int = 100,
    noise_formula: callable = lambda x: x / 2,
    max_noise_std: float = 0.2,
    copy_adata_var_info: bool = True,
    **kwargs,
):
    # Generate random delta vectors for each dimension
    span_adata = iterate_dimensions(
        latent_dims=embed.var["original_dim_id"].values,
        latent_min=embed.var["min"].values,
        latent_max=embed.var["max"].values,
        n_steps=n_steps,
        n_samples=n_samples,
    )

    # make baseline noise for each sample
    noise_std = noise_formula(embed.var["std"].values).clip(0, max_noise_std).reshape(1, -1)
    sample_wise_noises = np.random.randn(n_samples, embed.n_vars).astype(np.float32) * noise_std
    noise_vector = sample_wise_noises[span_adata.obs["sample_id"]]

    # make categorical covariates for each sample
    if model.adata_manager.get_state_registry(scvi.REGISTRY_KEYS.CAT_COVS_KEY):
        n_cats_per_key = model.adata_manager.get_state_registry(scvi.REGISTRY_KEYS.CAT_COVS_KEY).n_cats_per_key
        sample_wise_cats = np.stack([np.random.randint(0, n_cat, size=n_samples) for n_cat in n_cats_per_key], axis=1)
        cat_vector = sample_wise_cats[span_adata.obs["sample_id"]]
    else:
        cat_vector = None

    if model.adata_manager.get_state_registry(scvi.REGISTRY_KEYS.CONT_COVS_KEY):
        raise NotImplementedError("Interpretability of models with continuous covariates are not implemented yet.")

    # lib size
    lib_vector = np.ones(n_samples) * 1e4
    lib_vector = lib_vector[span_adata.obs["sample_id"]]

    # Control and effect latent data
    control_data = noise_vector
    effect_data = noise_vector + span_adata.X.A

    print("traversing latent ...")

    print(f"Input latent shape: control: {control_data.shape}, effect: {effect_data.shape}")
    # control and effect in mean parameter space
    control_mean_param = model.decode_latent_samples(
        control_data, lib=lib_vector, cat_key=cat_vector, cont_key=None, **kwargs
    )
    effect_mean_param = model.decode_latent_samples(
        effect_data, lib=lib_vector, cat_key=cat_vector, cont_key=None, **kwargs
    )
    print(f"Output mean param shape: control: {control_mean_param.shape}, effect: {effect_mean_param.shape}")

    if copy_adata_var_info:
        traverse_adata_var = model.adata.var.copy()
    else:
        traverse_adata_var = pd.DataFrame(index=model.adata.var_names)
    traverse_adata_var["original_order"] = np.arange(effect_mean_param.shape[1])

    traverse_adata = AnnData(
        X=effect_mean_param - control_mean_param,
        obs=span_adata.obs,
        var=traverse_adata_var,
    )
    traverse_adata.layers["control"] = control_mean_param
    traverse_adata.layers["effect"] = effect_mean_param
    traverse_adata.obsm["control_latent"] = control_data
    traverse_adata.obsm["effect_latent"] = effect_data
    if cat_vector is not None:
        traverse_adata.obsm["cat_covs"] = cat_vector
    traverse_adata.obs["lib_size"] = lib_vector

    return traverse_adata


def get_dimensions_of_traverse_data(traverse_adata: AnnData):
    # Get the number of latent dimensions, steps, samples, and vars
    n_latent = traverse_adata.obs["dim_id"].nunique()
    n_steps = traverse_adata.obs["step_id"].nunique()
    n_samples = traverse_adata.obs["sample_id"].nunique()
    n_vars = traverse_adata.n_vars
    return n_latent, n_steps, n_samples, n_vars


def traverse_latent(
    model: DRVI,
    embed: AnnData,
    n_steps=10 * 2,
    n_samples=100,
    copy_adata_var_info=True,
    **kwargs,
):
    if "original_dim_id" not in embed.var:
        raise ValueError(
            'Column "original_dim_id" not found in `embed.var`. Please run `set_latent_dimension_stats` to set vanished status.'
        )

    traverse_adata = make_traverse_adata(
        model=model,
        embed=embed,
        n_steps=n_steps,
        n_samples=n_samples,
        copy_adata_var_info=copy_adata_var_info,
        **kwargs,
    )

    # enrich traverse_adata with the additional info
    for col in ["title", "vanished", "order"]:
        mapping = dict(zip(embed.var["original_dim_id"].values, embed.var[col].values, strict=False))
        traverse_adata.obs[col] = traverse_adata.obs["dim_id"].map(mapping)

    return traverse_adata



from ._fields import (
    MerlinCategoricalJointObsField,
    MerlinCategoricalObsField,
    MerlinDataField,
    MerlinLayerField,
    MerlinNumericalJointObsField,
)

__all__ = [
    "MerlinCategoricalJointObsField",
    "MerlinCategoricalObsField",
    "MerlinLayerField",
    "MerlinNumericalJointObsField",
    "MerlinDataField",
]



import torch
from torch import nn
from torch.distributions import Normal, kl_divergence

# Standard, VaMP, GMM from Karin's CSI repo


class Prior(nn.Module):
    def __init__(self):
        super().__init__()

    def kl(self, qz):
        raise NotImplementedError()


class StandardPrior(Prior):
    def __init__(self):
        super().__init__()

    def kl(self, qz):
        # 1 x N
        assert isinstance(qz, Normal)
        return kl_divergence(qz, Normal(torch.zeros_like(qz.mean), torch.ones_like(qz.mean)))


class VampPrior(Prior):
    # Adapted from https://github.com/jmtomczak/intro_dgm/main/vaes/vae_priors_example.ipynb
    # K - components, I - inputs, L - latent, N - samples
    def __init__(
        self,
        n_components,
        encoder,
        model_input,
        trainable_keys=("x",),
        fixed_keys=(),
        input_type="scvi",
        preparation_function=None,
    ):
        super().__init__()

        self.encoder = encoder
        self.input_type = input_type
        self.preparation_function = preparation_function

        # pseudo inputs
        pi_aux_data = {}
        pi_tensor_data = {}
        for key in fixed_keys:
            if isinstance(model_input[key], torch.Tensor):
                pi_tensor_data[key] = torch.nn.Parameter(model_input[key], requires_grad=False)
            else:
                pi_aux_data[key] = model_input[key]
        for key in trainable_keys:
            pi_tensor_data[key] = torch.nn.Parameter(model_input[key], requires_grad=True)
            assert pi_tensor_data[key].shape[0] == n_components
        self.pi_aux_data = pi_aux_data
        self.pi_tensor_data = nn.ParameterDict(pi_tensor_data)

        # mixing weights
        self.w = torch.nn.Parameter(torch.zeros(n_components, 1, 1))  # K x 1 x 1

    def get_params(self):
        # u->encoder->mean, var
        original_mode = self.encoder.training
        self.encoder.train(False)
        if self.input_type == "scfemb":
            z = self.encoder({**self.pi_aux_data, **self.pi_tensor_data})
            output = z["qz_mean"], z["qz_var"]
        elif self.input_type == "scvi":
            x, args, kwargs = self.preparation_function({**self.pi_aux_data, **self.pi_tensor_data})
            q_m, q_v, latent = self.encoder(x, *args, **kwargs)
            output = q_m, q_v
        else:
            self.encoder.train(original_mode)
            raise NotImplementedError()
        self.encoder.train(original_mode)
        return output  # (K x L), (K x L)

    def log_prob(self, z):
        # Mixture of gaussian computed on K x N x L
        z = z.unsqueeze(0)  # 1 x N x L

        # u->encoder->mean, var
        m_p, v_p = self.get_params()  # (K x L), (K x L)
        m_p = m_p.unsqueeze(1)  # K x 1 x L
        v_p = v_p.unsqueeze(1)  # K x 1 x L

        # mixing probabilities
        w = torch.nn.functional.softmax(self.w, dim=0)  # K x 1 x 1

        # sum of log_p across components weighted by w
        log_prob = Normal(m_p, v_p.sqrt()).log_prob(z) + torch.log(w)  # K x N x L
        log_prob = torch.logsumexp(log_prob, dim=0, keepdim=False)  # N x L

        return log_prob  # N x L

    def kl(self, qz):
        assert isinstance(qz, Normal)
        z = qz.rsample()
        return qz.log_prob(z) - self.log_prob(z)

    def get_extra_state(self):
        return {
            "pi_aux_data": self.pi_aux_data,
            "input_type": self.input_type,
        }

    def set_extra_state(self, state):
        self.pi_aux_data = state["pi_aux_data"]
        self.input_type = state["input_type"]


class GaussianMixtureModelPrior(Prior):
    # Based on VampPrior class

    def __init__(
        self,
        n_components,
        n_latent,
        data=None,
        trainable_priors=True,
    ):
        # Do we need python 2 compatibility?
        super().__init__()

        if data is None:
            p_m = torch.rand(n_components, n_latent)  # K x L
            p_v = torch.ones(n_components, n_latent)  # K x L
        else:
            p_m = data[0]
            p_v = data[1]
        self.p_m = torch.nn.Parameter(p_m, requires_grad=trainable_priors)
        self.p_v = torch.nn.Parameter(p_v, requires_grad=trainable_priors)

        # mixing weights
        self.w = torch.nn.Parameter(torch.zeros(self.p_m.shape[0], 1, 1))  # K x 1 x 1

    def log_prob(self, z):
        # Mixture of gaussian computed on K x N x L
        z = z.unsqueeze(0)  # 1 x N x L

        m_p = self.p_m.unsqueeze(1)  # K x 1 x L
        v_p = self.p_v.unsqueeze(1)  # K x 1 x L

        # mixing probabilities
        w = torch.nn.functional.softmax(self.w, dim=0)  # K x 1 x 1

        # sum of log_p across components weighted by w
        log_prob = Normal(m_p, v_p.sqrt()).log_prob(z) + torch.log(w)  # K x N x L
        log_prob = torch.logsumexp(log_prob, dim=0, keepdim=False)  # N x L

        return log_prob  # N x L

    def kl(self, qz):
        assert isinstance(qz, Normal)
        z = qz.rsample()
        return qz.log_prob(z) - self.log_prob(z)



import merlin.dtypes
import merlin.schema
import pyarrow as pa


def read_first_row(path):
    """Read first row of a parquet file in a memory efficient way"""
    parquet_file = pa.parquet.ParquetFile(path)
    first_row = next(parquet_file.iter_batches(batch_size=1))
    df = pa.Table.from_batches([first_row])
    return df.to_pylist()[0]


def transfer_type_from_pyarrow(schema, first_row=None):
    """Transfer pyarrow schema to merlin schema."""
    if isinstance(schema, pa.Schema):
        fields = []
        for field in schema:
            coresponsing_first_row = None
            if first_row is not None:
                coresponsing_first_row = first_row[field.name]
            fields.append(transfer_type_from_pyarrow(field, first_row=coresponsing_first_row))
        return merlin.schema.Schema(fields)
    elif isinstance(schema, pa.Field):
        return merlin.schema.ColumnSchema(schema.name, **transfer_type_from_pyarrow(schema.type, first_row=first_row))
    elif isinstance(schema, pa.lib.ListType):
        return {
            "dtype": transfer_type_from_pyarrow(schema.value_type),
            "is_list": True,
            "is_ragged": False,
            "properties": {"value_count": {"max": len(first_row)}},
        }
    elif isinstance(schema, pa.DataType):
        if pa.float32().equals(schema):
            return {"dtype": merlin.dtypes.float32}
        elif pa.float64().equals(schema):
            return {"dtype": merlin.dtypes.float64}
        elif pa.int32().equals(schema):
            return {"dtype": merlin.dtypes.int32}
        elif pa.int64().equals(schema):
            return {"dtype": merlin.dtypes.int64}
        elif pa.bool_().equals(schema):
            return {"dtype": merlin.dtypes.boolean}
        else:
            raise ValueError(f"Unknown type for {schema}")
    else:
        raise ValueError(f"Unknown type for {schema}")



from ._latent import set_latent_dimension_stats
from .interpretability import (
    calculate_differential_vars,
    get_split_effects,
    iterate_on_top_differential_vars,
    traverse_latent,
)

__all__ = [
    "set_latent_dimension_stats",
    "traverse_latent",
    "calculate_differential_vars",
    "iterate_on_top_differential_vars",
    "get_split_effects",
]



from collections.abc import Callable, Iterable, Sequence
from typing import Literal

import numpy as np
import torch
from scvi import REGISTRY_KEYS
from scvi.module.base import BaseModuleClass, LossOutput, auto_move_data
from torch.distributions import Normal
from torch.utils.data import DataLoader

from drvi.nn_modules.embedding import MultiEmbedding
from drvi.nn_modules.layer.factory import LayerFactory
from drvi.nn_modules.noise_model import (
    LogNegativeBinomialNoiseModel,
    NegativeBinomialNoiseModel,
    NormalNoiseModel,
    PoissonNoiseModel,
)
from drvi.nn_modules.prior import GaussianMixtureModelPrior, StandardPrior, VampPrior
from drvi.scvi_tools_based.nn import DecoderDRVI, Encoder

TensorDict = dict[str, torch.Tensor]


class DRVIModule(BaseModuleClass):
    """
    Skeleton Variational auto-encoder model.

    Here we implement a basic version of scVI's underlying VAE :cite:p:`Lopez18`.
    This implementation is for instructional purposes only.

    Parameters
    ----------
    n_input
        Number of input genes
    n_latent
        Dimensionality of the latent space
    n_split_latent
        Number of splits in the latent space. -1 means split all dimensions (n_split_latent=n_latent).
    split_aggregation
        How to aggregate splits in the last layer of the decoder
    split_method
        How to make splits.
        "split" for splitting the latent space,
        "power" for transforming the latent space to n_split vectors of size n_latent
        "split_map" for splitting the latent space then map each to latent space using unique transformations
    decoder_reuse_weights
        Were to reuse the weights of the decoder layers when using splitting
        Possible values are 'everywhere', 'last', 'intermediate', 'nowhere'. Defaults to "everywhere"/
    encoder_dims
        Number of nodes in hidden layers of the encoder.
    decoder_dims
        Number of nodes in hidden layers of the decoder.
    n_cats_per_cov
        Number of categories for each categorical covariate
    n_continuous_cov
        Number of continuous covariates
    encode_covariates
        Whether to concatenate covariates to expression in encoder
    deeply_inject_covariates
        Whether to concatenate covariates into output of hidden layers in encoder/decoder. This option
        only applies when `n_layers` >= 1. The covariates are concatenated to the input of subsequent hidden layers.
    categorical_covariate_dims
        Emb dim of covariate keys if applicable
    covariate_modeling_strategy
        The strategy model takes to remove covariates
    use_batch_norm
        Whether to use batch norm in layers.
    affine_batch_norm
        Whether to use affine batch norm in layers.
    use_layer_norm
        Whether to use layer norm in layers.
    input_dropout_rate
        Dropout rate to apply to the input
    encoder_dropout_rate
        Dropout rate to apply to each of the encoder hidden layers
    decoder_dropout_rate
        Dropout rate to apply to each of the decoder hidden layers
    gene_likelihood
        gene likelihood model
    prior
        Prior model. defaults to normal.
    prior_init_dataloader
        Dataloader constructed to initialize the prior (or maintain in vamp).
    var_activation
        The activation function to ensure positivity of the variatinal distribution. Defaults to "exp".
    mean_activation
        The activation function at the end of mean encoder. Defaults to "identity".
        Possible values are "identity", "relu", "leaky_relu", "leaky_relu_{slope}", "elu", "elu_{min_vaule}".
    encoder_layer_factory
        A layer Factory instance for build encoder layers
    decoder_layer_factory
        A layer Factory instance for build decoder layers
    extra_encoder_kwargs
        Extra keyword arguments passed into encoder.
    extra_decoder_kwargs
        Extra keyword arguments passed into decoder.
    """

    def __init__(
        self,
        n_input: int,
        n_latent: int = 32,
        n_split_latent: int | None = -1,
        split_aggregation: Literal["sum", "logsumexp", "max"] = "logsumexp",
        split_method: Literal["split", "power", "split_map"] = "split_map",
        decoder_reuse_weights: Literal["everywhere", "last", "intermediate", "nowhere"] = "everywhere",
        encoder_dims: Sequence[int] = (128, 128),
        decoder_dims: Sequence[int] = (128, 128),
        n_cats_per_cov: Iterable[int] | None = (),
        n_continuous_cov: int = 0,
        encode_covariates: bool = False,
        deeply_inject_covariates: bool = False,
        categorical_covariate_dims: Sequence[int] = (),
        covariate_modeling_strategy: Literal[
            "one_hot",
            "emb",
            "emb_shared",
            "one_hot_linear",
            "emb_linear",
            "emb_shared_linear",
        ] = "one_hot",
        use_batch_norm: Literal["encoder", "decoder", "none", "both"] = "none",
        affine_batch_norm: Literal["encoder", "decoder", "none", "both"] = "both",
        use_layer_norm: Literal["encoder", "decoder", "none", "both"] = "both",
        fill_in_the_blanks_ratio: float = 0.0,
        input_dropout_rate: float = 0.0,
        encoder_dropout_rate: float = 0.1,
        decoder_dropout_rate: float = 0.0,
        gene_likelihood: Literal[
            "normal",
            "normal_v",
            "normal_sv",
            "poisson",
            "poisson_orig",
            "nb",
            "nb_sv",
            "nb_libnorm",
            "nb_loglib_rec",
            "nb_libnorm_loglib_rec",
            "nb_loglibnorm_all",
            "nb_orig",
            "nb_softmax",
            "nb_orig_libnorm",
            "pnb",
            "pnb_sv",
            "pnb_softmax",
        ] = "pnb_softmax",
        prior: Literal["normal", "gmm_x", "vamp_x"] = "normal",
        prior_init_dataloader: DataLoader | None = None,
        var_activation: Callable | Literal["exp", "pow2", "2sig"] | Callable = "exp",
        mean_activation: Callable | str = "identity",
        encoder_layer_factory: LayerFactory = None,
        decoder_layer_factory: LayerFactory = None,
        extra_encoder_kwargs: dict | None = None,
        extra_decoder_kwargs: dict | None = None,
    ):
        super().__init__()
        self.n_latent = n_latent
        if n_split_latent == -1:
            n_split_latent = n_latent
        self.n_split_latent = n_split_latent
        self.split_aggregation = split_aggregation
        self.latent_distribution = "normal"
        self.gene_likelihood = gene_likelihood

        self.encode_covariates = encode_covariates
        self.deeply_inject_covariates = deeply_inject_covariates

        self.gene_likelihood_module = self._construct_gene_likelihood_module(gene_likelihood)
        self.fill_in_the_blanks_ratio = fill_in_the_blanks_ratio

        use_batch_norm_encoder = use_batch_norm == "encoder" or use_batch_norm == "both"
        use_batch_norm_decoder = use_batch_norm == "decoder" or use_batch_norm == "both"
        affine_batch_norm_encoder = affine_batch_norm == "encoder" or affine_batch_norm == "both"
        affine_batch_norm_decoder = affine_batch_norm == "decoder" or affine_batch_norm == "both"
        use_layer_norm_encoder = use_layer_norm == "encoder" or use_layer_norm == "both"
        use_layer_norm_decoder = use_layer_norm == "decoder" or use_layer_norm == "both"

        assert covariate_modeling_strategy in [
            "one_hot",
            "emb",
            "emb_shared",
            "one_hot_linear",
            "emb_linear",
            "emb_shared_linear",
        ]
        if covariate_modeling_strategy in ["emb_shared", "emb_shared_linear"] and len(n_cats_per_cov) > 0:
            self.shared_covariate_emb = MultiEmbedding(
                n_cats_per_cov, categorical_covariate_dims, init_method="normal", max_norm=1.0
            )
        else:
            self.register_module("shared_covariate_emb", None)

        self.z_encoder = Encoder(
            n_input,
            n_latent,
            layers_dim=encoder_dims,
            input_dropout_rate=input_dropout_rate,
            dropout_rate=encoder_dropout_rate,
            n_cat_list=n_cats_per_cov if self.encode_covariates else [],
            n_continuous_cov=n_continuous_cov if self.encode_covariates else 0,
            inject_covariates=deeply_inject_covariates,
            use_batch_norm=use_batch_norm_encoder,
            affine_batch_norm=affine_batch_norm_encoder,
            use_layer_norm=use_layer_norm_encoder,
            var_activation=var_activation,
            mean_activation=mean_activation,
            layer_factory=encoder_layer_factory,
            covariate_modeling_strategy=covariate_modeling_strategy,
            categorical_covariate_dims=categorical_covariate_dims if self.encode_covariates else [],
            **(extra_encoder_kwargs or {}),
        )
        self.decoder = DecoderDRVI(
            n_latent,
            n_input,
            n_split=n_split_latent,
            split_aggregation=split_aggregation,
            split_method=split_method,
            reuse_weights=decoder_reuse_weights,
            gene_likelihood_module=self.gene_likelihood_module,
            layers_dim=decoder_dims,
            dropout_rate=decoder_dropout_rate,
            n_cat_list=n_cats_per_cov,
            n_continuous_cov=n_continuous_cov,
            inject_covariates=deeply_inject_covariates,
            use_batch_norm=use_batch_norm_decoder,
            affine_batch_norm=affine_batch_norm_decoder,
            use_layer_norm=use_layer_norm_decoder,
            layer_factory=decoder_layer_factory,
            covariate_modeling_strategy=covariate_modeling_strategy,
            categorical_covariate_dims=categorical_covariate_dims,
            **(extra_decoder_kwargs or {}),
        )

        self.prior = self._construct_prior(prior, prior_init_dataloader)
        self.fully_deterministic = False

    def _construct_gene_likelihood_module(self, gene_likelihood):
        if gene_likelihood == "normal":
            return NormalNoiseModel(model_var="fixed=1")
        elif gene_likelihood == "normal_v":
            return NormalNoiseModel(model_var="dynamic")
        elif gene_likelihood == "normal_sv":
            return NormalNoiseModel(model_var="feature")
        elif gene_likelihood == "poisson":
            return PoissonNoiseModel(mean_transformation="exp", library_normalization="none")
        elif gene_likelihood == "poisson_orig":
            return PoissonNoiseModel(mean_transformation="softmax", library_normalization="none")
        elif gene_likelihood in ["nb", "nb_sv"]:
            return NegativeBinomialNoiseModel(dispersion="feature", library_normalization="none")
        elif gene_likelihood in ["nb_libnorm"]:
            return NegativeBinomialNoiseModel(dispersion="feature", library_normalization="x_lib")
        elif gene_likelihood in ["nb_loglib_rec"]:
            return NegativeBinomialNoiseModel(dispersion="feature", library_normalization="x_loglib")
        elif gene_likelihood in ["nb_libnorm_loglib_rec"]:
            return NegativeBinomialNoiseModel(dispersion="feature", library_normalization="div_lib_x_loglib")
        elif gene_likelihood in ["nb_loglibnorm_all"]:
            return NegativeBinomialNoiseModel(dispersion="feature", library_normalization="x_loglib_all")
        elif gene_likelihood in ["nb_orig", "nb_softmax"]:
            return NegativeBinomialNoiseModel(
                dispersion="feature", mean_transformation="softmax", library_normalization="none"
            )
        elif gene_likelihood == "nb_orig_libnorm":
            return NegativeBinomialNoiseModel(
                dispersion="feature", mean_transformation="softmax", library_normalization="x_lib"
            )
        elif gene_likelihood in ["pnb", "pnb_sv"]:
            return LogNegativeBinomialNoiseModel(dispersion="feature", library_normalization="none")
        elif gene_likelihood in ["pnb_softmax"]:
            return LogNegativeBinomialNoiseModel(
                dispersion="feature", mean_transformation="softmax", library_normalization="none"
            )
        else:
            raise NotImplementedError()

    def _construct_prior(self, prior, prior_init_dataloader=None):
        if prior == "normal":
            return StandardPrior()
        elif prior.startswith("gmm_"):
            n_components = int(prior.split("_")[1])
            if prior_init_dataloader is not None:
                inference_output = self.inference(**self._get_inference_input(next(iter(prior_init_dataloader))))
                init_data = inference_output["qz_m"], inference_output["qz_v"]
            else:
                init_data = None
            return GaussianMixtureModelPrior(n_components, self.n_latent, data=init_data)
        elif prior.startswith("vamp_"):
            n_components = int(prior.split("_")[1])
            if prior_init_dataloader is not None:

                def preparation_function(prepared_input):
                    x = prepared_input["encoder_input"]
                    args = []
                    kwargs = {"cat_full_tensor": prepared_input["cat_full_tensor"]}
                    return x, args, kwargs

                model_input = self._input_pre_processing(**self._get_inference_input(next(iter(prior_init_dataloader))))
            else:
                raise ValueError("VaMP prior needs input samples as pseudo-inputs.")
            return VampPrior(
                n_components,
                self.z_encoder,
                model_input,
                input_type="scvi",
                trainable_keys=("encoder_input",),
                fixed_keys=("cat_full_tensor",),
                preparation_function=preparation_function,
            )
        else:
            raise NotImplementedError()

    def _get_inference_input(self, tensors):
        """Parse the dictionary to get appropriate args"""
        x = tensors[REGISTRY_KEYS.X_KEY]

        cont_covs = tensors.get(REGISTRY_KEYS.CONT_COVS_KEY)
        cat_covs = tensors.get(REGISTRY_KEYS.CAT_COVS_KEY)

        input_dict = {"x": x, "cont_covs": cont_covs, "cat_covs": cat_covs}
        return input_dict

    def _input_pre_processing(self, x, cont_covs=None, cat_covs=None):
        # log the input to the variational distribution for numerical stability
        x_, gene_likelihood_additional_info = self.gene_likelihood_module.initial_transformation(x)
        # Note: this is different from scvi implementation of library size that is log transformed
        # All our noise models accept non-normalized library to work
        library = x.sum(1)

        encoder_input = x_

        return {
            "encoder_input": encoder_input,
            "cat_full_tensor": cat_covs if self.encode_covariates else None,
            "cont_full_tensor": cont_covs if self.encode_covariates else None,
            "library": library,
            "gene_likelihood_additional_info": gene_likelihood_additional_info,
        }

    @auto_move_data
    def inference(self, x, cont_covs=None, cat_covs=None):
        """
        High level inference method.

        Runs the inference (encoder) model.
        """
        pre_processed_input = self._input_pre_processing(x, cont_covs, cat_covs).copy()
        x_ = pre_processed_input["encoder_input"]

        # Mask if needed
        if self.fill_in_the_blanks_ratio > 0.0 and self.training:
            assert cont_covs is None  # We do not consider cont_cov here
            x_mask = torch.where(torch.rand_like(x_) >= self.fill_in_the_blanks_ratio, 1.0, 0.0)
            x_ = x_ * x_mask
            # TODO: check this:
            # x_ = x_ * x_mask / x_mask.mean(dim=1, keepdim=True)
        else:
            x_mask = None

        # Prepare shared emb
        if self.shared_covariate_emb is not None and self.encode_covariates:
            pre_processed_input["cat_full_tensor"] = self.shared_covariate_emb(
                pre_processed_input["cat_full_tensor"].int()
            )

        # get variational parameters via the encoder networks
        qz_m, qz_v, z = self.z_encoder(
            x_,
            cat_full_tensor=pre_processed_input["cat_full_tensor"],
            cont_full_tensor=pre_processed_input["cont_full_tensor"],
        )

        outputs = {
            "z": z,
            "qz_m": qz_m,
            "qz_v": qz_v,
            "library": pre_processed_input["library"],
            "x_mask": x_mask,
            "gene_likelihood_additional_info": pre_processed_input["gene_likelihood_additional_info"],
        }
        return outputs

    def _get_generative_input(self, tensors, inference_outputs):
        z = inference_outputs["z"]
        if self.fully_deterministic:
            z = inference_outputs["qz_m"]
        library = inference_outputs["library"]
        gene_likelihood_additional_info = inference_outputs["gene_likelihood_additional_info"]

        cont_covs = tensors.get(REGISTRY_KEYS.CONT_COVS_KEY)
        cat_covs = tensors.get(REGISTRY_KEYS.CAT_COVS_KEY)

        input_dict = {
            "z": z,
            "library": library,
            "gene_likelihood_additional_info": gene_likelihood_additional_info,
            "cont_covs": cont_covs,
            "cat_covs": cat_covs,
        }
        return input_dict

    @auto_move_data
    def generative(self, z, library, gene_likelihood_additional_info, cont_covs=None, cat_covs=None):
        """Runs the generative model."""
        if self.shared_covariate_emb is not None:
            cat_covs = self.shared_covariate_emb(cat_covs.int())
        # form the likelihood
        px, params, original_params = self.decoder(
            z,
            cat_full_tensor=cat_covs,
            cont_full_tensor=cont_covs,
            library=library,
            gene_likelihood_additional_info=gene_likelihood_additional_info,
        )

        return {
            "px": px,
            "params": params,
            "original_params": original_params,
        }

    def loss(
        self,
        tensors,
        inference_outputs,
        generative_outputs,
        kl_weight: float = 1.0,
    ):
        """Loss function."""
        x = tensors[REGISTRY_KEYS.X_KEY]
        x_mask = inference_outputs["x_mask"]
        qz_m = inference_outputs["qz_m"]
        qz_v = inference_outputs["qz_v"]
        px = generative_outputs["px"]

        kl_divergence_z = self.prior.kl(Normal(qz_m, torch.sqrt(qz_v))).sum(dim=1)
        if self.fill_in_the_blanks_ratio > 0.0 and self.training:
            reconst_loss = -(px.log_prob(x) * (1 - x_mask)).sum(dim=-1)
        else:
            reconst_loss = -px.log_prob(x).sum(dim=-1)
        # For MSE this should be equivalent (in terms of backward gradients) to:
        # reconst_loss = torch.nn.GaussianNLLLoss(reduction='none')(x, px.loc, px.scale ** 2).sum(dim=-1)
        assert kl_divergence_z.shape == reconst_loss.shape

        kl_local_for_warmup = kl_divergence_z
        kl_local_no_warmup = 0.0

        weighted_kl_local = kl_weight * kl_local_for_warmup + kl_local_no_warmup

        loss = torch.mean(reconst_loss + weighted_kl_local)

        kl_local = {"kl_divergence_z": kl_divergence_z.sum()}
        return LossOutput(
            loss=loss,
            reconstruction_loss=reconst_loss,
            kl_local=kl_local,
            extra_metrics={
                "mse": torch.nn.functional.mse_loss(x, px.mean, reduction="none").sum(dim=1).mean(dim=0),
            },
        )

    @torch.no_grad()
    def sample(
        self,
        tensors,
        n_samples=1,
        library_size=1,
    ) -> torch.Tensor:
        # Note: Not tested
        r"""
        Generate observation samples from the posterior predictive distribution.

        The posterior predictive distribution is written as :math:`p(\hat{x} \mid x)`.

        Parameters
        ----------
        tensors
            Tensors dict
        n_samples
            Number of required samples for each cell
        library_size
            Library size to scale scamples to

        Returns
        -------
        x_new
            tensor with shape (n_cells, n_genes, n_samples)
        """
        inference_kwargs = dict(n_samples=n_samples)  # noqa: C408
        (
            _,
            generative_outputs,
        ) = self.forward(
            tensors,
            inference_kwargs=inference_kwargs,
            compute_loss=False,
        )

        dist = generative_outputs["px"]

        if n_samples > 1:
            exprs = dist.sample().permute([1, 2, 0])  # Shape : (n_cells_batch, n_genes, n_samples)
        else:
            exprs = dist.sample()

        return exprs.cpu()

    @torch.no_grad()
    @auto_move_data
    def marginal_ll(self, tensors: TensorDict, n_mc_samples: int):
        # Note: Not tested
        """Marginal ll."""
        sample_batch = tensors[REGISTRY_KEYS.X_KEY]

        to_sum = torch.zeros(sample_batch.size()[0], n_mc_samples)

        for i in range(n_mc_samples):
            # Distribution parameters and sampled variables
            inference_outputs, _, losses = self.forward(tensors)
            qz_m = inference_outputs["qz_m"]
            qz_v = inference_outputs["qz_v"]
            z = inference_outputs["z"]

            # Reconstruction Loss
            reconst_loss = losses.dict_sum(losses.reconstruction_loss)

            # Log-probabilities

            p_z = Normal(torch.zeros_like(qz_m), torch.ones_like(qz_v)).log_prob(z).sum(dim=-1)
            p_x_zl = -reconst_loss

            to_sum[:, i] = p_z + p_x_zl

        batch_log_lkl = torch.logsumexp(to_sum, dim=-1) - np.log(n_mc_samples)
        log_lkl = torch.sum(batch_log_lkl).item()
        return log_lkl



from collections.abc import Sequence
from typing import Literal

import anndata as ad
import numpy as np
import scanpy as sc
from anndata import AnnData
from matplotlib import pyplot as plt

from drvi.utils.plotting import cmap


def make_balanced_subsample(adata, col, min_count=10):
    n_sample_per_cond = adata.obs[col].value_counts().min()
    balanced_sample_index = (
        adata.obs.groupby(col)
        .sample(n=max(min_count, n_sample_per_cond), random_state=0, replace=n_sample_per_cond < min_count)
        .index
    )
    adata = adata[balanced_sample_index].copy()
    return adata


def plot_latent_dimension_stats(
    embed: AnnData,
    figsize: tuple[int, int] = (5, 3),
    log_scale: bool | Literal["try"] = "try",
    ncols: int = 5,
    columns: Sequence[str] = ("reconstruction_effect", "max_value", "mean", "std"),
    titles: dict[str, str] = None,
    remove_vanished: bool = False,
    show: bool = True,
):
    """
    Plot the statistics of latent dimensions.

    Args:
    - embed (AnnData): The annotated data object containing the latent dimensions.
    - figsize (Tuple[int, int], optional): The size of the figure (width, height). Default is (5, 3).
    - log_scale (Union[bool, Literal['try']], optional): Whether to use a log scale for the y-axis. If 'try', the log scale is used if the minimum value is greater than 0. Default is 'try'.
    - ncols (int, optional): The maximum number of columns in the subplot grid. Default is 5.
    - columns (Sequence[str], optional): The columns to plot from the `embed` object. Default is ('reconstruction_effect', 'max_value', 'mean', 'std').
    - titles (Dict[str, str], optional): The titles for each column in the plot.
    - show (bool, optional): Whether to display the plot. If False, figure is returned. Default is True.

    Returns
    -------
    - plt (matplotlib.pyplot module): The matplotlib.pyplot module if show is False.
    """
    if titles is None:
        titles = {
            "reconstruction_effect": "Reconstruction effect",
            "max_value": "Max value",
            "mean": "Mean",
            "std": "Standard Deviation",
        }
    nrows = int(np.ceil(len(columns) / ncols))
    if nrows == 1:
        ncols = len(columns)

    fig, axes = plt.subplots(
        nrows, ncols, figsize=(figsize[0] * ncols, figsize[1] * nrows), sharey=False, sharex=False, squeeze=False
    )

    # Iterate through columns and plot the data
    for ax, col in zip(axes.flatten(), columns, strict=False):
        df = embed.var
        if remove_vanished:
            df = df.query("vanished == False")
        df = df.sort_values("order")
        ranks = df["order"]
        x = df[col]

        ax.plot(ranks, x, linestyle="-", color="grey", label="Line")  # Solid line plot
        for vanished_status_to_plot in [True, False]:
            indices = df["vanished"] == vanished_status_to_plot
            ax.plot(
                ranks[indices],
                x[indices],
                "o",
                markersize=3,
                color="black" if vanished_status_to_plot else "blue",
                label="Data Points",
            )

        # Adding labels and title
        ax.set_xlabel("Rank based on Explanation Share")
        ax.set_ylabel(titles[col] if col in titles else col)
        if isinstance(log_scale, str):
            if log_scale == "try":
                if x.min() > 0:
                    ax.set_yscale("log")
        else:
            if log_scale:
                ax.set_yscale("log")

        # Removing the legend
        ax.legend().remove()

        # Adding grid
        ax.grid(axis="x")

    if not remove_vanished:
        # Create custom legend entries
        handles = []
        for vanished_status_to_plot in [False, True]:
            color = "black" if vanished_status_to_plot else "blue"
            label = "Vanished" if vanished_status_to_plot else "Non-vanished"
            handles.append(
                plt.Line2D([0], [0], marker="o", color="w", markerfacecolor=color, markersize=5, label=label)
            )

        # Add the legend to the first subplot or the entire figure
        fig.legend(
            handles=handles,
            labels=[handle.get_label() for handle in handles],
            loc="center left",
            bbox_to_anchor=(1, 0.5),
            title=None,
        )

    for ax in axes.flatten()[len(columns) :]:
        fig.delaxes(ax)

    plt.tight_layout()
    if show:
        plt.show()
    else:
        return fig


def plot_latent_dims_in_umap(
    embed,
    title_col: str = "title",
    additional_columns=(),
    max_cells_to_plot: int | None = None,
    order_col="order",
    dim_subset: Sequence[str] | None = None,
    directional: bool = False,
    remove_vanished: bool = True,
    rearrange_titles: bool = True,
    color_bar_rescale_ratio: float = 1.0,
    show: bool = True,
    **kwargs,
):
    """
    Plot the latent dimensions of a UMAP embedding.

    Args:
        embed (AnnData): Annotated data object containing the UMAP embedding.
        title_col (str, optional): Name of the column in `embed.var` to use as titles for each dimension.
                                   If None, default titles will be used.
        additional_columns (tuple, optional): Additional columns to plot alongside the latent dimensions.
        max_cells_to_plot (int, optional): Maximum number of cells to plot. If the number of cells in `embed`
                                           is greater than `max_cells_to_plot`, a subsample will be taken.
        order_col (str, optional): The column in the `embed.var` DataFrame to use for ordering the dimensions. Ignored if `dim_subset` is provided.
        dim_subset (Sequence[str], optional): The subset of dimensions to plot. Defaults to None.
        directional (bool, optional): Consider + and - directions as two separate dimensions. Default is False.
        remove_vanished (bool, optional): Whether to remove the vanished dimensions from the plot. Default is True.
        rearrange_titles (bool, optional): Whether to rearrange the titles to bottom right of the plot. Default is True.
        color_bar_rescale_ratio (float, optional): The ratio to rescale the height of colorbars. Default is 1.0.
        show (bool, optional): Whether to display the plot. If False, the plot is returned. Default is True.
        **kwargs: Additional keyword arguments to be passed to `sc.pl.umap`.

    Returns
    -------
        matplotlib.figure.Figure: The UMAP plot figure if show is False.
    """
    if order_col not in embed.var:
        raise ValueError(
            f'Column "{order_col}" not found in `embed.var`. Please run `set_latent_dimension_stats` to set order.'
        )
    if remove_vanished and "vanished" not in embed.var:
        raise ValueError(
            'Column "vanished" not found in `embed.var`. Please run `set_latent_dimension_stats` to set vanished status.'
        )

    if max_cells_to_plot is not None and embed.n_obs > max_cells_to_plot:
        embed = sc.pp.subsample(embed, n_obs=max_cells_to_plot, copy=True)

    if directional:
        embed_pos = embed.copy()
        if "layer" in kwargs:
            embed_pos.X = embed_pos.layers[kwargs["layer"]]
            del kwargs["layer"]
        embed_neg = embed_pos.copy()
        embed_neg.X = -embed_neg.X
        embed_neg.var["min"], embed_neg.var["max"] = -embed_neg.var["max"], -embed_neg.var["min"]

        embed_pos.var["_direction"] = "+"
        embed_neg.var["_direction"] = "-"
        embed_pos.var[title_col] = embed_pos.var[title_col] + "+"
        embed_neg.var[title_col] = embed_neg.var[title_col] + "-"
        if embed.var[order_col].dtype in [np.int32, np.int64, np.float32, np.float64]:
            embed_pos.var[order_col] = embed_pos.var[order_col] + 1e-8
            embed_neg.var[order_col] = embed_neg.var[order_col] - 1e-8
        else:
            embed_pos.var[order_col] = embed_pos.var[order_col].astype(str) + "+"
            embed_neg.var[order_col] = embed_neg.var[order_col].astype(str) + "-"

        embed = ad.concat([embed_pos, embed_neg], axis=1, join="inner", merge="first")
        embed.var.reset_index(drop=True, inplace=True)

    tmp_df = embed.var.sort_values(order_col)
    if remove_vanished:
        tmp_df = tmp_df.query("vanished == False")
    if dim_subset:
        tmp_df = tmp_df.set_index(title_col).loc[dim_subset].reset_index()
    cols_to_show = tmp_df.index if title_col is None else tmp_df[title_col]
    if additional_columns:
        cols_to_show = list(cols_to_show) + list(additional_columns)

    kwargs = {
        **dict(  # noqa: C408
            frameon=False,
            cmap=cmap.saturated_red_blue_cmap if not directional else cmap.saturated_sky_cmap,
            vmin=list(np.minimum(tmp_df["min"].values, -1)),
            vcenter=0,
            vmax=list(np.maximum(tmp_df["max"].values, +1)),
        ),
        **kwargs,
    }
    fig = sc.pl.umap(embed, gene_symbols=title_col, color=cols_to_show, return_fig=True, **kwargs)
    for i, ax in enumerate(fig.axes[1 : 2 * len(tmp_df) : 2]):
        assert hasattr(ax, "_colorbar")
        pos = ax.get_position()
        new_pos = [pos.x0, pos.y0, pos.width, pos.height * color_bar_rescale_ratio]
        ax.set_position(new_pos)

        if directional:
            direction = tmp_df["_direction"].iloc[i]
            if direction == "-":
                ax.invert_yaxis()
                labels = -ax.get_yticks()
                if all(x == int(x) for x in labels):
                    labels = [int(x) for x in labels]
                ax.set_yticklabels(labels)
    if rearrange_titles:
        for ax in fig.axes:
            ax.text(0.935, 0.05, ax.get_title(), size=15, ha="left", color="black", rotation=90, transform=ax.transAxes)
            ax.set_title("")

    if show:
        plt.show()
    else:
        return fig


def plot_latent_dims_in_heatmap(
    embed: AnnData,
    categorical_column: str,
    title_col: str | None = "title",
    sort_by_categorical: bool = False,
    make_balanced: bool = True,
    order_col: str | None = "order",
    remove_vanished: bool = True,
    figsize: tuple[int, int] | None = None,
    show: bool = True,
    **kwargs,
):
    """
    Plot the latent dimensions in a heatmap.

    Parameters
    ----------
        embed (AnnData): The annotated data object containing the latent dimensions.
        categorical_column (str): The column in the `embed.obs` DataFrame that represents the categorical variable.
        title_col (str, optional): The column in the `embed.var` DataFrame to use as the title for each dimension. Defaults to None.
        sort_by_categorical (bool, optional): Whether to sort the dimensions based on the categorical variable. Defaults to True.
        make_balanced (bool, optional): Whether to make a balanced subsample of the data based on the categorical variable. Defaults to True.
        order_col (str, optional): The column in the `embed.var` DataFrame to use for ordering the dimensions. Discarded if sort_by_categorical is set. Defaults to 'order'.
        remove_vanished (bool, optional): Whether to remove the vanished dimensions from the plot. Default is True.
        figsize (Tuple[int, int], optional): The size of the figure. Defaults to None.
        show (bool, optional): Whether to show the plot. Defaults to True.
        **kwargs: Additional keyword arguments to be passed to the `sc.pl.heatmap` function.

    Returns
    -------
        plot if Show is not False.
    """
    if order_col is not None and order_col not in embed.var:
        raise ValueError(
            f'Column "{order_col}" not found in `embed.var`. Please run `set_latent_dimension_stats` to set order.'
        )
    if remove_vanished:
        if "vanished" not in embed.var:
            raise ValueError(
                'Column "vanished" not found in `embed.var`. Please run `set_latent_dimension_stats` to set vanished status.'
            )
        embed = embed[:, ~embed.var["vanished"]]

    if make_balanced:
        embed = make_balanced_subsample(embed, categorical_column)

    if sort_by_categorical:
        dim_order = np.abs(embed.X).argmax(axis=0).argsort().tolist()
    elif order_col is None:
        dim_order = np.arange(embed.n_vars)
    else:
        dim_order = embed.var[order_col].argsort().tolist()

    if title_col is None:
        vars_to_show = embed.var.iloc[dim_order].index
    else:
        vars_to_show = embed.var.iloc[dim_order][title_col]

    if figsize is None:
        figsize = (10, len(embed.obs[categorical_column].unique()) / 6)

    kwargs = {
        **dict(  # noqa: C408
            vcenter=0,
            cmap=cmap.saturated_red_blue_cmap,
            dendrogram=False,
        ),
        **kwargs,
    }

    return sc.pl.heatmap(
        embed,
        vars_to_show,
        categorical_column,
        gene_symbols=title_col,
        figsize=figsize,
        show_gene_labels=True,
        show=show,
        **kwargs,
    )



from importlib.metadata import version

from . import model, nn_modules, scvi_tools_based, utils

__version__ = version("drvi-py")

__all__ = [
    "__version__",
    "model",
    "utils",
    "nn_modules",
    "scvi_tools_based",
]



import numpy as np
from anndata import AnnData

from drvi.model import DRVI


def set_latent_dimension_stats(
    model: DRVI,
    embed: AnnData,
    inplace: bool = True,
    vanished_threshold=0.1,
):
    """
    Set the latent dimension statistics of a DRVI model into var of an embedding anndata.

    Parameters
    ----------
    model
        DRVI model object.
    embed
        latent representation of the model.
    inplace
        Whether to modify the input AnnData object or return a new one.
    """
    if not inplace:
        embed = embed.copy()

    if "original_dim_id" not in embed.var:
        embed.var["original_dim_id"] = np.arange(embed.var.shape[0])

    embed.var["reconstruction_effect"] = 0
    embed.var["reconstruction_effect"].loc[embed.var.sort_values("original_dim_id").index] = (
        model.get_reconstruction_effect_of_each_split()
    )
    embed.var["order"] = (-embed.var["reconstruction_effect"]).argsort().argsort()

    embed.var["max_value"] = np.abs(embed.X).max(axis=0)
    embed.var["mean"] = embed.X.mean(axis=0)
    embed.var["min"] = embed.X.min(axis=0)
    embed.var["max"] = embed.X.max(axis=0)
    embed.var["std"] = np.abs(embed.X).std(axis=0)

    embed.var["title"] = "DR " + (1 + embed.var["order"]).astype(str)
    embed.var["vanished"] = embed.var["max_value"] < vanished_threshold

    if not inplace:
        return embed



import logging

import torch
from anndata import AnnData
from scvi import REGISTRY_KEYS
from scvi.data._constants import _MODEL_NAME_KEY, _SETUP_ARGS_KEY
from scvi.model._utils import parse_device_args
from scvi.model.base import BaseModelClass
from scvi.model.base._archesmixin import ArchesMixin, _get_loaded_data, _initialize_model, _validate_var_names
from torch import nn

from drvi.scvi_tools_based.nn import FCLayers

logger = logging.getLogger(__name__)


class DRVIArchesMixin(ArchesMixin):
    """Universal scArches implementation."""

    @classmethod
    def load_query_data(
        cls,
        adata: AnnData,
        reference_model: str | BaseModelClass,
        inplace_subset_query_vars: bool = False,
        accelerator: str = "auto",
        device: int | str = "auto",
        unfrozen: bool = False,
        freeze_dropout: bool = False,
        freeze_shared_emb: bool = True,
        freeze_encoder: bool = True,
        freeze_decoder: bool = True,
        reset_encoder: bool = False,
        reset_decoder: bool = False,
        freeze_batchnorm_encoder: bool = True,
        freeze_batchnorm_decoder: bool = False,
    ):
        """Online update of a reference model with scArches algorithm :cite:p:`Lotfollahi21`.

        Parameters
        ----------
        adata
            AnnData organized in the same way as data used to train model.
            It is not necessary to run setup_anndata,
            as AnnData is validated against the ``registry``.
        reference_model
            Either an already instantiated model of the same class, or a path to
            saved outputs for reference model.
        inplace_subset_query_vars
            Whether to subset and rearrange query vars inplace based on vars used to
            train reference model.
        %(param_accelerator)s
        %(param_device)s
        unfrozen
            Override all other freeze options for a fully unfrozen model
        freeze_dropout
            Whether to freeze dropout during training
        freeze_shared_emb
            Whether to freeze shared embeddings if any
        freeze_encoder
            Whether to freeze encoder
        freeze_decoder
            Whether to freeze decoder
        freeze_batchnorm_encoder
            Whether to freeze encoder batchnorms' weight and bias during transfer
        freeze_batchnorm_decoder
            Whether to freeze decoder batchnorms' weight and bias during transfer
        """
        _, _, device = parse_device_args(
            accelerator=accelerator,
            devices=device,
            return_device="torch",
            validate_single_device=True,
        )

        attr_dict, var_names, load_state_dict = _get_loaded_data(reference_model, device=device)

        if inplace_subset_query_vars:
            logger.debug("Subsetting query vars to reference vars.")
            adata._inplace_subset_var(var_names)
        _validate_var_names(adata, var_names)

        registry = attr_dict.pop("registry_")
        if _MODEL_NAME_KEY in registry and registry[_MODEL_NAME_KEY] != cls.__name__:
            raise ValueError("It appears you are loading a model from a different class.")

        if _SETUP_ARGS_KEY not in registry:
            raise ValueError("Saved model does not contain original setup inputs. Cannot load the original setup.")

        cls.setup_anndata(
            adata,
            source_registry=registry,
            extend_categories=True,
            allow_missing_labels=True,
            **registry[_SETUP_ARGS_KEY],
        )

        model = _initialize_model(cls, adata, attr_dict)
        adata_manager = model.get_anndata_manager(adata, required=True)

        if REGISTRY_KEYS.CAT_COVS_KEY in adata_manager.data_registry:
            previous_n_cats_per_cov = registry["field_registries"][REGISTRY_KEYS.CAT_COVS_KEY]["state_registry"][
                "n_cats_per_key"
            ]
            n_cats_per_cov = model.adata_manager.get_state_registry(REGISTRY_KEYS.CAT_COVS_KEY).n_cats_per_key
        else:
            previous_n_cats_per_cov = None
            n_cats_per_cov = None

        model.to_device(device)

        reloaded_tensor_keys = []
        # model tweaking
        new_state_dict = model.module.state_dict()
        for key, load_ten in load_state_dict.items():
            new_ten = new_state_dict[key]
            if reset_decoder and "decoder" in key:
                assert not freeze_decoder
                print(f"Resetting {key} in decoder")
                load_state_dict[key] = new_ten
                reloaded_tensor_keys.append(key)
                continue
            if reset_encoder and "z_encoder" in key:
                assert not freeze_encoder
                print(f"Resetting {key} in encoder")
                load_state_dict[key] = new_ten
                reloaded_tensor_keys.append(key)
                continue
            if new_ten.size() == load_ten.size():
                continue
            # new categoricals changed size
            else:
                print(f"Resizing {key} from {load_ten.size()} to {new_ten.size()}")
                if "emb_list" in key:
                    # Extend embeddings along dim 0 (n_emb)
                    dim_diff = new_ten.size()[0] - load_ten.size()[0]
                    assert new_ten.size()[1] == load_ten.size()[1]
                    fixed_ten = torch.cat([load_ten, new_ten[-dim_diff:, :]], dim=0)
                    load_state_dict[key] = fixed_ten
                    reloaded_tensor_keys.append(key)
                else:
                    if new_ten.dim() == load_ten.dim() == 2:
                        # 2D tensors
                        # Extend linear layers along dim 1 (input)
                        assert (
                            sum(n_cats_per_cov) - sum(previous_n_cats_per_cov)
                            == new_ten.size()[-1] - load_ten.size()[-1]
                        )
                        assert new_ten.size()[0] == load_ten.size()[0]
                        # Keep data flow (normal nodes) as is
                        cum_n_cat_old = new_ten.size()[1] - sum(n_cats_per_cov)
                        cum_n_cat_new = cum_n_cat_old
                        fixed_ten = [load_ten[:, :cum_n_cat_old]] if cum_n_cat_old > 0 else []
                        # Iterate and get old covariates deom load_ten and new ones from new_ten
                        for n_cat_new, n_cat_old in zip(n_cats_per_cov, previous_n_cats_per_cov, strict=False):
                            fixed_ten.append(load_ten[:, cum_n_cat_old : cum_n_cat_old + n_cat_old])
                            if n_cat_new > n_cat_old:
                                fixed_ten.append(new_ten[:, cum_n_cat_new : cum_n_cat_new + n_cat_new - n_cat_old])
                            cum_n_cat_old += n_cat_old
                            cum_n_cat_new += n_cat_new
                        # Concat and set as init tensor
                        fixed_ten = torch.cat([t.to(device) for t in fixed_ten], dim=1)
                        load_state_dict[key] = fixed_ten
                        reloaded_tensor_keys.append(key)
                    elif new_ten.dim() == load_ten.dim() == 3:
                        # 3D tensors
                        # extend weight of stacked linears along dim 1 (input)
                        assert (
                            sum(n_cats_per_cov) - sum(previous_n_cats_per_cov) == new_ten.size()[1] - load_ten.size()[1]
                        )
                        assert new_ten.size()[0] == load_ten.size()[0]
                        assert new_ten.size()[2] == load_ten.size()[2]
                        # Keep data flow (normal nodes) as is
                        cum_n_cat_old = new_ten.size()[1] - sum(n_cats_per_cov)
                        cum_n_cat_new = cum_n_cat_old
                        fixed_ten = [load_ten[:, :cum_n_cat_old, :]] if cum_n_cat_old > 0 else []
                        # Iterate and get old covariates deom load_ten and new ones from new_ten
                        for n_cat_new, n_cat_old in zip(n_cats_per_cov, previous_n_cats_per_cov, strict=False):
                            fixed_ten.append(load_ten[:, cum_n_cat_old : cum_n_cat_old + n_cat_old, :])
                            if n_cat_new > n_cat_old:
                                fixed_ten.append(new_ten[:, cum_n_cat_new : cum_n_cat_new + n_cat_new - n_cat_old])
                            cum_n_cat_old += n_cat_old
                            cum_n_cat_new += n_cat_new
                        # Concat and set as init tensor
                        fixed_ten = torch.cat([t.to(device) for t in fixed_ten], dim=1)
                        load_state_dict[key] = fixed_ten
                        reloaded_tensor_keys.append(key)
                    else:
                        raise NotImplementedError()

        model.module.load_state_dict(load_state_dict)
        model.module.eval()

        # Make hooks to make gradients zero while training
        _set_params_online_update(
            model.module,
            reloaded_tensor_keys,
            unfrozen=unfrozen,
            previous_n_cats_per_cov=previous_n_cats_per_cov,
            n_cats_per_cov=n_cats_per_cov,
            freeze_dropout=freeze_dropout,
            freeze_shared_emb=freeze_shared_emb,
            freeze_encoder=freeze_encoder,
            freeze_decoder=freeze_decoder,
            freeze_batchnorm_encoder=freeze_batchnorm_encoder,
            freeze_batchnorm_decoder=freeze_batchnorm_decoder,
        )
        model.is_trained_ = False

        return model


def _set_params_online_update(
    module,
    reloaded_tensor_keys,
    unfrozen,
    previous_n_cats_per_cov,
    n_cats_per_cov,
    freeze_dropout,
    freeze_shared_emb,
    freeze_encoder,
    freeze_decoder,
    freeze_batchnorm_encoder,
    freeze_batchnorm_decoder,
):
    """Freeze parts of network for scArches."""
    # do nothing if unfrozen
    if unfrozen:
        return

    if freeze_shared_emb:
        if hasattr(module, "shared_covariate_emb") and module.shared_covariate_emb is not None:
            print(f"Freezing top {previous_n_cats_per_cov} items in Shared Emb.")
            assert tuple(module.shared_covariate_emb.num_embeddings) == tuple(n_cats_per_cov)
            module.shared_covariate_emb.freeze_top_embs(previous_n_cats_per_cov)

    def requires_grad(key):
        if not freeze_decoder and "decoder" in key:
            return True
        if not freeze_encoder and "z_encoder" in key:
            return True
        # Extended tensors should have gradients but we put some gradients to zero by online hooks
        if key in reloaded_tensor_keys:
            return True
        return False

    # General changes
    for key, par in module.named_parameters():
        if requires_grad(key):
            par.requires_grad = True
        else:
            print(f"Freezing key {key}")
            par.requires_grad = False

    # specific changes
    for key, mod in module.named_modules():
        # skip over protected modules
        if isinstance(mod, FCLayers):
            freeze_fc_layers = ("decoder" in key and freeze_decoder) or ("z_encoder" in key and freeze_encoder)
            # This will make requires_frad for layers after one_hot to True
            if freeze_fc_layers:
                print(f"Setting hooks for {key}")
                mod.set_online_update_hooks(previous_n_cats_per_cov, n_cats_per_cov)
        elif isinstance(mod, nn.Dropout):
            if freeze_dropout:
                print(f"Freezing dropout {key}")
                mod.p = 0
        elif isinstance(mod, nn.BatchNorm1d | nn.LayerNorm):
            assert hasattr(mod, "freeze")
            freeze_batchnorm = ("decoder" in key and freeze_batchnorm_decoder) or (
                "z_encoder" in key and freeze_batchnorm_encoder
            )
            if freeze_batchnorm:
                print(f"Freezing normalization layer {key}")
                mod.freeze(freeze_batchnorm)
                mod.momentum = 0



import lightning.pytorch as pl

from drvi.scvi_tools_based.merlin_data._data_loader import MerlinTransformedDataLoader
from drvi.scvi_tools_based.merlin_data._data_manager import MerlinDataManager


class MerlinDataSplitter(pl.LightningDataModule):
    """Creates data loaders for MerlinData object given the MerlinDataManager.

    Parameters
    ----------
    adata_manager
        MerlinDataManager object that has been created by setup_merlin_data.
    """

    data_loader_cls = MerlinTransformedDataLoader

    def __init__(
        self,
        adata_manager: MerlinDataManager,
        **kwargs,
    ):
        super().__init__()
        self.adata_manager = adata_manager
        # We some usual inputs
        for key in ["train_size", "validation_size", "shuffle_set_split", "pin_memory"]:
            # print("Discarding key:", key)
            kwargs.pop(key, None)
        self.kwargs = kwargs

    def setup(self, stage: str | None = None):
        self.val_idx = None
        self.train_idx = None
        self.test_idx = None

    def _get_dataloader(self, split, shuffle=False, **kwargs):
        parts_per_chunk = {
            "train": 8,
            "val": 1,
            "test": 1,
        }[split]
        return self.data_loader_cls(
            self.adata_manager.get_dataset(split),
            mapping=self.adata_manager.get_fields_schema_mapping(),
            shuffle=shuffle,
            parts_per_chunk=parts_per_chunk,
            **kwargs,
        )

    def train_dataloader(self):
        """Create train data loader."""
        return self._get_dataloader("train", shuffle=True, **self.kwargs)

    def val_dataloader(self):
        """Create validation data loader."""
        return self._get_dataloader("val", shuffle=False, **self.kwargs)

    def test_dataloader(self):
        """Create test data loader."""
        return self._get_dataloader("test", shuffle=False, **self.kwargs)



import numpy as np
import pandas as pd
import torch

from drvi.nn_modules.embedding import FeatureEmbedding


class TestFeatureEmbedding:
    def make_test_data(self):
        feature_table = pd.DataFrame(
            [
                ["g1", "g1", "GE"],
                ["g2", "g2", "GE"],
                ["g3", "g3", "GE"],
                ["p1", "g1", "PK"],
                ["p2", "g1", "PK"],
                ["p3", "g4", "PK"],
            ],
            columns=["id", "nearest_gene", "feature_type"],
        )
        feature_table_row_extended = pd.DataFrame(
            [
                ["g3", "g3", "GE"],
                ["g4", "g4", "GE"],
                ["p4", "g9", "PK"],
                ["p6", "g3", "PK"],
            ],
            columns=["id", "nearest_gene", "feature_type"],
        )
        feature_table_full_extended = pd.DataFrame(
            [
                ["g3", "g3", "GE", "NEW_FEATURE_1"],
                ["p1", "g1", "KK", "NEW_FEATURE_2"],
                ["p100", "g100", "PK", "NEW_FEATURE_2"],
            ],
            columns=["id", "nearest_gene", "feature_type", "new_feature"],
        )

        return feature_table, feature_table_row_extended, feature_table_full_extended

    def test_feature_embedding_simple(self):
        sample_df = self.make_test_data()[0]
        feature_embedding = FeatureEmbedding.from_pandas_dataframe(sample_df, [2, 2, 1])
        accessed_data = feature_embedding(
            np.array(
                [
                    ["g3", "g3", "GE"],
                    ["p1", "g3", "PK"],
                ]
            )
        )
        assert ((accessed_data[0, 2:4] - accessed_data[1, 2:4]) ** 2).sum() < 1e-8  # g3 is same

    def test_feature_embedding_cache(self):
        sample_df = self.make_test_data()[0]
        feature_embedding = FeatureEmbedding.from_pandas_dataframe(sample_df, [2, 2, 1])
        accessed_data = feature_embedding(
            np.array(
                [
                    ["g3", "g3", "GE"],
                    ["p1", "g3", "PK"],
                ]
            ),
            index_cache_key="x",
        )
        assert ((feature_embedding(None, index_cache_key="x") - accessed_data) ** 2).sum() < 1e-8

    def test_feature_embedding_row_extension(self):
        sample_df, row_extension_df, _ = self.make_test_data()
        union_df = pd.concat([sample_df, row_extension_df])

        feature_embedding = FeatureEmbedding.from_pandas_dataframe(sample_df, [2, 2, 1])
        feature_embedding_extended = FeatureEmbedding.from_pandas_dataframe(union_df, [2, 2, 1])

        key_array = np.array(
            [
                ["g1", "g3", "GE"],
                ["p4", "g9", "PK"],
            ]
        )
        accessed_data_pre_transfer = feature_embedding_extended(key_array)
        feature_embedding_extended.load_weights_from_trained_module(feature_embedding)
        accessed_data_post_transfer = feature_embedding_extended(key_array)

        # First row should come from feature_embedding (overwritten)
        assert ((feature_embedding(key_array[0:1, :]) - accessed_data_post_transfer[0:1, :]) ** 2).sum() < 1e-8
        # In second row, 'p4', 'g9' should not change
        assert ((accessed_data_pre_transfer[1, :4] - accessed_data_post_transfer[1, :4]) ** 2).sum() < 1e-8

    def test_feature_embedding_full_extension(self):
        sample_df, _, full_extension_df = self.make_test_data()
        union_df = pd.concat([sample_df, full_extension_df])

        feature_embedding = FeatureEmbedding.from_pandas_dataframe(sample_df, [2, 2, 1])
        feature_embedding_extended = FeatureEmbedding.from_pandas_dataframe(union_df, [2, 2, 2, 2])

        key_array = np.array(
            [
                ["g1", "g3", "GE", "NEW_FEATURE_1"],
                ["p100", "g100", "KK", "NEW_FEATURE_2"],
            ]
        )
        accessed_data_pre_transfer = feature_embedding_extended(key_array)
        feature_embedding_extended.load_weights_from_trained_module(feature_embedding)
        accessed_data_post_transfer = feature_embedding_extended(key_array)

        # 'g1', 'g3', 'GE (dim1)' should come from feature_embedding (overwritten)
        assert ((feature_embedding(key_array[0:1, :3]) - accessed_data_post_transfer[0:1, :-3]) ** 2).sum() < 1e-8
        # Last 3 dims (1 for 'feature_type' and 2 for 'new_feature') should not change (new)
        assert ((accessed_data_pre_transfer[:, -3:] - accessed_data_post_transfer[:, -3:]) ** 2).sum() < 1e-8
        # Second row should not change (new)
        assert ((accessed_data_pre_transfer[1, :] - accessed_data_post_transfer[1, :]) ** 2).sum() < 1e-8

    def test_feature_embedding_freeze_on_extension(self):
        sample_df, _, full_extension_df = self.make_test_data()
        union_df = pd.concat([sample_df, full_extension_df])

        feature_embedding = FeatureEmbedding.from_pandas_dataframe(sample_df, [2, 2, 1])
        feature_embedding_extended = FeatureEmbedding.from_pandas_dataframe(union_df, [2, 2, 2, 2])

        key_array = np.array(
            [
                ["g1", "g3", "GE", "NEW_FEATURE_1"],
                ["p100", "g100", "KK", "NEW_FEATURE_2"],
            ]
        )
        feature_embedding_extended.load_weights_from_trained_module(feature_embedding, freeze_old=True)
        # grad(x^2/2) = 2 -> grad = data if not freeze
        loss = sum([(p**2 / 2).sum() for _, p in feature_embedding_extended.named_parameters()])
        loss.backward()
        with torch.no_grad():
            for _name, p in list(feature_embedding_extended.named_parameters()):
                p.data -= p.grad  # everything that is not freeze should become zero

        accessed_data_post_transfer = feature_embedding_extended(key_array)
        # 'g1', 'g3', 'GE (dim1)' should come from feature_embedding (overwritten)
        assert ((feature_embedding(key_array[0:1, :3]) - accessed_data_post_transfer[0:1, :-3]) ** 2).sum() < 1e-8
        # Last 3 dims should become zero
        assert ((accessed_data_post_transfer[:, -3:]) ** 2).sum() < 1e-8
        # Second row should become zero
        assert ((accessed_data_post_transfer[1, :]) ** 2).sum() < 1e-8



from typing import Literal

import numpy as np
import pandas as pd
import scipy
from anndata import AnnData

from drvi.model import DRVI
from drvi.utils.tools.interpretability._latent_traverse import get_dimensions_of_traverse_data, traverse_latent


def find_differential_effects(
    traverse_adata: AnnData,
    method: Literal["max_possible", "min_possible"] = "max_possible",
    key_added: str = "effect",
    add_to_counts: float = 0.1,
    relax_max_by: float = 0.0,
):
    assert method in ["max_possible", "min_possible"]

    # Reorder the traverse_adata to original order
    original_traverse_adata = traverse_adata
    traverse_adata = traverse_adata[traverse_adata.obs.sort_values(["original_order"]).index].copy()
    traverse_adata = traverse_adata[:, traverse_adata.var.sort_values(["original_order"]).index].copy()

    # Get the number of latent dimensions, steps, samples, and vars
    n_latent, n_steps, n_samples, n_vars = get_dimensions_of_traverse_data(traverse_adata)

    # Get the dim_id and span values
    span_values = traverse_adata.obs["span_value"].values.reshape(n_latent, n_steps, n_samples)
    assert np.allclose(span_values, span_values.max(axis=-1, keepdims=True))
    span_values = span_values[:, :, 0]  # n_latent x n_steps
    dim_ids = traverse_adata.obs["dim_id"].values.reshape(n_latent, n_steps, n_samples)[:, 0, 0]  # n_latent

    # Get the output mean parameters in 4D format
    control_mean_param = traverse_adata.layers["control"].reshape(n_latent, n_steps, n_samples, n_vars)
    effect_mean_param = traverse_adata.layers["effect"].reshape(n_latent, n_steps, n_samples, n_vars)

    # Helper functions
    average_over_samples = lambda x: x.mean(axis=2)
    add_eps_in_count_space = lambda x: scipy.special.logsumexp(
        np.stack([x, np.log(add_to_counts) * np.ones_like(x)]), axis=0
    )
    find_relative_effect = lambda x, baseline: (
        scipy.special.logsumexp(np.stack([x, np.log(add_to_counts) * np.ones_like(x), baseline]), axis=0) - baseline
    )

    # Find DE for each sample and average over samples
    if method == "max_possible":
        diff_considering_small_values = average_over_samples(
            add_eps_in_count_space(effect_mean_param) - add_eps_in_count_space(control_mean_param)
        )  # n_latent x n_steps x n_vars
    elif method == "min_possible":
        reduce_dims = (
            1,
            2,
        )
        max_of_two = np.maximum(
            effect_mean_param.max(axis=reduce_dims, keepdims=True),
            control_mean_param.max(axis=reduce_dims, keepdims=True),
        )  # n_latent x 1 x n_samples, n_vars
        max_cumulative_possible_all = scipy.special.logsumexp(
            max_of_two, axis=0, keepdims=True
        )  # 1 x 1 x n_samples, n_vars
        max_cumulative_possible_other_dims = (
            np.log(np.exp(max_cumulative_possible_all) - np.exp(max_of_two)) - relax_max_by
        )  # n_latent x 1 x n_samples, n_vars
        max_cumulative_possible_other_dims = max_cumulative_possible_other_dims + np.zeros_like(
            effect_mean_param
        )  # n_latent x n_steps x n_samples, n_vars
        normalized_effect_mean_param = find_relative_effect(
            effect_mean_param, max_cumulative_possible_other_dims
        )  # n_latent x n_steps x n_samples, n_vars
        normalized_control_mean_param = find_relative_effect(
            control_mean_param, max_cumulative_possible_other_dims
        )  # n_latent x n_steps x n_samples, n_vars
        diff_considering_small_values = average_over_samples(
            normalized_effect_mean_param - normalized_control_mean_param
        )  # n_latent x n_steps x n_vars
    else:
        raise NotImplementedError()

    original_traverse_adata.uns[f"{key_added}_traverse_effect_stepwise"] = diff_considering_small_values

    # Find DE vars in positive and negative directions
    for effect_sign in ["pos", "neg"]:
        mask = np.where(span_values >= 0, 1, 0) if effect_sign == "pos" else np.where(span_values <= 0, 1, 0)
        max_effect = np.max(np.expand_dims(mask, axis=-1) * diff_considering_small_values, axis=1)  # n_latent x n_vars
        max_effect = pd.DataFrame(max_effect.T, index=traverse_adata.var_names, columns=dim_ids)
        original_traverse_adata.varm[f"{key_added}_traverse_effect_{effect_sign}"] = max_effect.loc[
            original_traverse_adata.var_names
        ].copy()
        original_traverse_adata.uns[f"{key_added}_traverse_effect_{effect_sign}_dim_ids"] = (
            original_traverse_adata.varm[f"{key_added}_traverse_effect_{effect_sign}"].columns.values
        )
        original_traverse_adata.varm[
            f"{key_added}_traverse_effect_{effect_sign}"
        ].columns = original_traverse_adata.varm[f"{key_added}_traverse_effect_{effect_sign}"].columns.astype(str)


def combine_differential_effects(
    traverse_adata: AnnData,
    keys: list[str],
    key_added: str,
    combine_function: callable,
):
    # Reorder the traverse_adata to original order
    original_traverse_adata = traverse_adata
    traverse_adata = traverse_adata[traverse_adata.obs.sort_values(["original_order"]).index].copy()
    traverse_adata = traverse_adata[:, traverse_adata.var.sort_values(["original_order"]).index].copy()

    # Get the number of latent dimensions, steps, samples, and vars
    n_latent, n_steps, n_samples, n_vars = get_dimensions_of_traverse_data(traverse_adata)

    # Get the dim_id and span values
    span_values = traverse_adata.obs["span_value"].values.reshape(n_latent, n_steps, n_samples)
    assert np.allclose(span_values, span_values.max(axis=-1, keepdims=True))
    span_values = span_values[:, :, 0]  # n_latent x n_steps
    dim_ids = traverse_adata.obs["dim_id"].values.reshape(n_latent, n_steps, n_samples)[:, 0, 0]  # n_latent

    # Combine effects
    combined_traverse_effect_stepwise = combine_function(
        *[traverse_adata.uns[f"{key}_traverse_effect_stepwise"] for key in keys]
    )

    original_traverse_adata.uns[f"{key_added}_traverse_effect_stepwise"] = combined_traverse_effect_stepwise

    # Find DE vars in positive and negative directions
    for effect_sign in ["pos", "neg"]:
        mask = np.where(span_values >= 0, 1, 0) if effect_sign == "pos" else np.where(span_values <= 0, 1, 0)
        max_effect = np.max(
            np.expand_dims(mask, axis=-1) * combined_traverse_effect_stepwise, axis=1
        )  # n_latent x n_vars
        max_effect = pd.DataFrame(max_effect.T, index=traverse_adata.var_names, columns=dim_ids)
        original_traverse_adata.varm[f"{key_added}_traverse_effect_{effect_sign}"] = max_effect.loc[
            original_traverse_adata.var_names
        ].copy()
        original_traverse_adata.uns[f"{key_added}_traverse_effect_{effect_sign}_dim_ids"] = (
            original_traverse_adata.varm[f"{key_added}_traverse_effect_{effect_sign}"].columns.values
        )
        original_traverse_adata.varm[
            f"{key_added}_traverse_effect_{effect_sign}"
        ].columns = original_traverse_adata.varm[f"{key_added}_traverse_effect_{effect_sign}"].columns.astype(str)


def calculate_differential_vars(traverse_adata: AnnData, **kwargs):
    print("Finding differential variables per latent dimension ...")
    find_differential_effects(traverse_adata, method="max_possible", key_added="max_possible", **kwargs)
    find_differential_effects(traverse_adata, method="min_possible", key_added="min_possible", **kwargs)

    def combine_function(min_possible, max_possible):
        # min_possible and max_possible dimensions: n_latent x n_steps x n_vars
        keep = (max_possible >= 1.0) & (
            (max_possible > max_possible.max(axis=(0, 1), keepdims=True) / 2)
            | (min_possible > min_possible.max(axis=(0, 1), keepdims=True) / 10)
        )
        score = np.where(keep, max_possible * min_possible, 0)
        return score

    # Combine scores with product
    combine_differential_effects(
        traverse_adata,
        keys=["min_possible", "max_possible"],
        key_added="combined_score",
        combine_function=combine_function,
    )


def get_split_effects(
    model: DRVI,
    embed: AnnData,
    n_steps=10 * 2,
    n_samples=100,
    traverse_kwargs=None,
    de_kwargs=None,
):
    """
    Get the split effect of a latent dimension.

    Parameters
    ----------
    model
        DRVI model object.
    embed
        latent representation of the model.
    n_steps
        Number of steps.
    n_samples
        Number of samples.
    traverse_kwargs
        Additional arguments passed to `traverse_the_latent`.
    kwargs
        Additional arguments passed to `find_differential_vars`.
    """
    if traverse_kwargs is None:
        traverse_kwargs = {}
    if de_kwargs is None:
        de_kwargs = {}

    traverse_adata = traverse_latent(model, embed, n_steps=n_steps, n_samples=n_samples, **traverse_kwargs)
    calculate_differential_vars(traverse_adata, **de_kwargs)

    return traverse_adata


def iterate_on_top_differential_vars(
    traverse_adata: AnnData,
    key: str,
    title_col: str = "title",
    order_col: str = "order",
    gene_symbols: str | None = None,
    score_threshold: float = 0.0,
):
    """
    Make an iterator of the top differential variables per latent dimension.

    Parameters
    ----------
    traverse_adata
        Anndata object with the split effects.
    key
        Key to the differential variables in traverse_adata.
    title_col
        Title columns defining title of the dimensions.
    order_col
        Order column that defines latent dimension orders.
    gene_symbols
        Column with the gene symbols in traverse_adata.var.
    score_threshold
        A threshold to filter the differential variables.
    """
    df_pos = traverse_adata.varm[f"{key}_traverse_effect_pos"].copy()
    df_neg = traverse_adata.varm[f"{key}_traverse_effect_neg"].copy()

    if gene_symbols is not None:
        gene_name_mapping = dict(zip(traverse_adata.var.index, traverse_adata.var[gene_symbols], strict=False))
    else:
        gene_name_mapping = dict(zip(traverse_adata.var.index, traverse_adata.var.index, strict=False))
    df_pos.index = df_pos.index.map(gene_name_mapping)
    df_neg.index = df_neg.index.map(gene_name_mapping)

    de_info = dict(
        **{
            (str(k) + "+"): v.sort_values(ascending=False).where(lambda x: x > score_threshold).dropna()
            for k, v in df_pos.to_dict(orient="series").items()
        },
        **{
            (str(k) + "-"): v.sort_values(ascending=False).where(lambda x: x > score_threshold).dropna()
            for k, v in df_neg.to_dict(orient="series").items()
        },
    )

    return [
        (f"{row[title_col]}{direction}", de_info[f"{row['dim_id']}{direction}"])
        for i, row in traverse_adata.obs[["dim_id", order_col, title_col]]
        .drop_duplicates()
        .sort_values(order_col)
        .iterrows()
        for direction in ["-", "+"]
        if len(de_info[f"{row['dim_id']}{direction}"]) > 0
    ]



import numpy as np
import pandas as pd

from drvi.nn_modules.encodig import FeatureOneHotEncoding


class TestFeatureEmbedding:
    def make_test_data(self):
        sample_df = pd.DataFrame(
            [
                ["g1", "g1", "GE"],
                ["g2", "g2", "GE"],
                ["g3", "g3", "GE"],
                ["p1", "g1", "PK"],
                ["p2", "g1", "PK"],
                ["p3", "g4", "PK"],
            ],
            columns=["id", "nearest_gene", "feature_type"],
        )
        return sample_df

    def test_feature_encoding_simple(self):
        sample_df = self.make_test_data()
        feature_encoding = FeatureOneHotEncoding.from_pandas_dataframe(sample_df)
        print(feature_encoding)
        accessed_data = feature_encoding(
            np.array(
                [
                    ["g3", "g3", "GE"],
                    ["g3", "g3", "PK"],
                    ["g3", "g3", "GE"],
                ]
            ),
            index_cache_key="x",
        )
        assert ((accessed_data[0] - accessed_data[1]) ** 2).sum() >= 1
        assert ((accessed_data[0] - accessed_data[2]) ** 2).sum() < 1e-8
